import copy
import json

from sdcclient._common import _SdcCommon
from sdcclient.monitor.dashboard_converters import convert_dashboard_between_versions, \
    convert_scope_string_to_expression

PANEL_VISUALIZATION_TIMECHART = "advancedTimechart"
PANEL_VISUALIZATION_NUMBER = "advancedNumber"


class DashboardsClientV3(_SdcCommon):
    def __init__(self, token="", sdc_url='https://app.sysdigcloud.com', ssl_verify=True, custom_headers=None):
        super(DashboardsClientV3, self).__init__(token, sdc_url, ssl_verify, custom_headers)
        self.product = "SDC"
        self._dashboards_api_version = 'v3'
        self._dashboards_api_endpoint = '/api/{}/dashboards'.format(self._dashboards_api_version)
        self._default_dashboards_api_endpoint = '/api/{}/dashboards/templates'.format(self._dashboards_api_version)

    def get_views_list(self):
        res = self.http.get(self.url + self._default_dashboards_api_endpoint, headers=self.hdrs,
                            verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]
        return [True, res.json()]

    def get_view(self, name):
        gvres = self.get_views_list()
        if gvres[0] is False:
            return gvres

        vlist = gvres[1]['dashboardTemplates']

        id = None

        for v in vlist:
            if v['name'] == name:
                id = v['dashboardId']
                break

        if not id:
            return [False, 'view ' + name + ' not found']

        res = self.http.get(self.url + self._default_dashboards_api_endpoint + '/' + id, headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def get_dashboards(self, light=True):
        """
        Return the list of dashboards available under the given user account.
        This includes the dashboards created by the user and the ones shared with her by other users.
        Since every user has multiple tokens -- one per team they are assigned to -- and dashboards are scoped per
        team, in order to download all the dashboards from a user account, all the tokens must be specified.

        Args:
            light (bool): If it's true, only a small portion of information will be retrieved, otherwise all
                          the information from the dashboards will be retrieved.
                          If the retrieved dashboards are going to be restored in the future, then this
                          parameter must be false.

        Returns:
            The list of the dashboards visible for the given token.

        See Also:
            :py:meth:`~.DashboardsClientV3.create_dashboard_with_configuration`

        Examples:
            >>> ok, res = client.get_dashboards(light=False)
            >>> for dashboard in res["dashboards"]:
            >>>     print(dashboard["name"])
        """
        params = {
            "light": "true" if light else "false"
        }
        res = self.http.get(self.url + self._dashboards_api_endpoint, params=params,
                            headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def update_dashboard(self, dashboard_data):
        '''**Description**
            Updates dashboard with provided in data. Please note that the dictionary will require a valid ID and version field to work as expected.

        **Success Return Value**
            A dictionary containing the updated dashboard data.

        **Example**
            `examples/dashboard_basic_crud.py <https://github.com/draios/python-sdc-client/blob/master/examples/dashboard_basic_crud.py>`_
        '''
        res = self.http.put(self.url + self._dashboards_api_endpoint + "/" + str(dashboard_data['id']),
                            headers=self.hdrs, verify=self.ssl_verify, data=json.dumps({'dashboard': dashboard_data}))
        return self._request_result(res)

    def find_dashboard_by(self, name=None):
        '''**Description**
            Finds dashboards with the specified name. You can then delete the dashboard (with :func:`~SdcClient.delete_dashboard`) or edit panels (with :func:`~SdcClient.add_dashboard_panel` and :func:`~SdcClient.remove_dashboard_panel`)

        **Arguments**
            - **name**: the name of the dashboards to find.

        **Success Return Value**
            A list of dictionaries of dashboards matching the specified name.

        **Example**
            `examples/dashboard.py <https://github.com/draios/python-sdc-client/blob/master/examples/dashboard.py>`_
        '''
        res = self.get_dashboards()
        if res[0] is False:
            return res
        else:
            def filter_fn(configuration):
                return configuration['name'] == name

            def create_item(configuration):
                return {'dashboard': configuration}

            dashboards = list(map(create_item, list(filter(filter_fn, res[1]['dashboards']))))
            return [True, dashboards]

    def create_dashboard_with_configuration(self, configuration):
        # Remove id and version properties if already set
        configuration_clone = copy.deepcopy(configuration)
        if 'id' in configuration_clone:
            del configuration_clone['id']
        if 'version' in configuration_clone:
            del configuration_clone['version']

        res = self.http.post(self.url + self._dashboards_api_endpoint, headers=self.hdrs,
                             data=json.dumps({'dashboard': configuration_clone}),
                             verify=self.ssl_verify)
        return self._request_result(res)

    def create_dashboard(self, name):
        '''
        **Description**
            Creates an empty dashboard. You can then add panels by using ``add_dashboard_panel``.

        **Arguments**
            - **name**: the name of the dashboard that will be created.

        **Success Return Value**
            A dictionary showing the details of the new dashboard.

        **Example**
            `examples/dashboard.py <https://github.com/draios/python-sdc-client/blob/master/examples/dashboard.py>`_
        '''
        dashboard_configuration = {
            'name': name,
            'schema': 3,
            'widgets': [],
            'eventsOverlaySettings': {
                'filterNotificationsUserInputFilter': ''
            },
            'layout': [],
            'panels': [],
        }

        #
        # Create the new dashboard
        #
        res = self.http.post(self.url + self._dashboards_api_endpoint, headers=self.hdrs,
                             data=json.dumps({'dashboard': dashboard_configuration}),
                             verify=self.ssl_verify)
        return self._request_result(res)

    # TODO COVER
    def add_dashboard_panel(self, dashboard, panel_name, visualization, query):
        dboard = copy.deepcopy(dashboard)
        new_panel_id = dboard["panels"][-1]["id"] + 1
        new_panel = {
            "id": new_panel_id,
            "type": visualization,
            "name": panel_name,
            "description": "",
            "advancedQueries": [
                {
                    "enabled": True,
                    "displayInfo": {
                        "displayName": "",
                        "timeSeriesDisplayNameTemplate": "",
                        "type": "lines"
                    },
                    "format": {
                        "unit": "%",
                        "inputFormat": "0-100",
                        "displayFormat": "auto",
                        "decimals": None,
                        "yAxis": "auto"
                    },
                    "query": query
                }
            ]
        }
        new_layout = {
            "panelId": new_panel_id,
            "x": 0,
            # Hackish way to position a panel, the API doesn't provide auto-position
            "y": len(dboard["panels"]) * 12 + 12,
            "w": 12,
            "h": 6,
        }

        if visualization == PANEL_VISUALIZATION_TIMECHART:
            new_panel["axesConfiguration"] = {
                "bottom": {
                    "enabled": True
                },
                "left": {
                    "enabled": True,
                    "displayName": None,
                    "unit": "auto",
                    "displayFormat": "auto",
                    "decimals": None,
                    "minValue": 0,
                    "maxValue": None,
                    "minInputFormat": "0-100",
                    "maxInputFormat": "0-100",
                    "scale": "linear"
                },
                "right": {
                    "enabled": True,
                    "displayName": None,
                    "unit": "auto",
                    "displayFormat": "auto",
                    "decimals": None,
                    "minValue": 0,
                    "maxValue": None,
                    "minInputFormat": "1",
                    "maxInputFormat": "1",
                    "scale": "linear"
                }
            }
            new_panel["legendConfiguration"] = {
                "enabled": True,
                "position": "right",
                "layout": "table",
                "showCurrent": True
            }
        if visualization == PANEL_VISUALIZATION_NUMBER:
            new_panel["numberThresholds"] = {
                "values": [],
                "base": {
                    "severity": "none",
                    "displayText": "",
                }
            }

        dboard["panels"].append(new_panel)
        dboard["layout"].append(new_layout)

        return self.update_dashboard(dboard)

    # TODO COVER
    def remove_dashboard_panel(self, dashboard, panel_id):
        dboard = copy.deepcopy(dashboard)
        dboard["panels"] = [panel for panel in dboard["panels"] if panel["id"] != panel_id]
        dboard["layout"] = [layout for layout in dboard["layout"] if layout["panelId"] != panel_id]

        return self.update_dashboard(dboard)

    def create_dashboard_from_template(self, dashboard_name, template, scope=None, shared=False, public=False):
        if scope is not None:
            if not isinstance(scope, list) and not isinstance(scope, str):
                return [False, 'Invalid scope format: Expected a list, a string or None']
        else:
            scope = []

        #
        # Clean up the dashboard we retireved so it's ready to be pushed
        #
        template['id'] = None
        template['version'] = None
        template['schema'] = 3
        template['name'] = dashboard_name
        template['shared'] = shared
        template['public'] = public
        template['publicToken'] = None

        # default dashboards don't have eventsOverlaySettings property
        # make sure to add the default set if the template doesn't include it
        if 'eventsOverlaySettings' not in template or not template['eventsOverlaySettings']:
            template['eventsOverlaySettings'] = {
                'filterNotificationsUserInputFilter': ''
            }

        # set dashboard scope to the specific parameter
        template['scopeExpressionList'] = []
        if isinstance(scope, list):
            for s in scope:
                ok, converted_scope = convert_scope_string_to_expression(s)
                if not ok:
                    return ok, converted_scope
                template['scopeExpressionList'].append(converted_scope[0])
        elif isinstance(scope, str):
            ok, converted_scope = convert_scope_string_to_expression(scope)
            if not ok:
                return ok, converted_scope
            template['scopeExpressionList'] = converted_scope

        # NOTE: Individual panels might override the dashboard scope, the override will NOT be reset
        if 'widgets' in template and template['widgets'] is not None:
            for chart in template['widgets']:
                if 'overrideScope' not in chart:
                    chart['overrideScope'] = False

                if not chart['overrideScope']:
                    # patch frontend bug to hide scope override warning even when it's not really overridden
                    chart['scope'] = scope

                if chart['showAs'] != 'map':
                    # if chart scope is equal to dashboard scope, set it as non override
                    chart_scope = chart['scope'] if 'scope' in chart else None
                    chart['overrideScope'] = chart_scope != scope
                else:
                    # topology panels must override the scope
                    chart['overrideScope'] = True

        #
        # Create the new dashboard
        #
        res = self.http.post(self.url + self._dashboards_api_endpoint, headers=self.hdrs,
                             data=json.dumps({'dashboard': template}), verify=self.ssl_verify)

        return self._request_result(res)

    def create_dashboard_from_file(self, dashboard_name, filename, filter=None, shared=False, public=False):
        '''
        **Description**
            Create a new dasboard using a dashboard template saved to disk. See :func:`~SdcClient.save_dashboard_to_file` to use the file to create a dashboard (usefl to create and restore backups).

            The file can contain the following JSON formats:
            1. dashboard object in the format of an array element returned by :func:`~SdcClient.get_dashboards`
            2. JSON object with the following properties:
                * version: dashboards API version (e.g. 'v2')
                * dashboard: dashboard object in the format of an array element returned by :func:`~SdcClient.get_dashboards`

        **Arguments**
            - **dashboard_name**: the name of the dashboard that will be created.
            - **filename**: name of a file containing a JSON object
            - **filter**: a boolean expression combining Sysdig Monitor segmentation criteria defines what the new dasboard will be applied to. For example: *kubernetes.namespace.name='production' and container.image='nginx'*.
            - **shared**: if set to True, the new dashboard will be a shared one.
            - **public**: if set to True, the new dashboard will be shared with public token.

        **Success Return Value**
            A dictionary showing the details of the new dashboard.

        **Example**
            `examples/dashboard_save_load.py <https://github.com/draios/python-sdc-client/blob/master/examples/dashboard_save_load.py>`_
        '''
        #
        # Load the Dashboard
        #
        with open(filename) as data_file:
            loaded_object = json.load(data_file)

        #
        # Handle old files
        #
        if 'dashboard' not in loaded_object:
            loaded_object = {
                'version': f'v{loaded_object["schema"]}',
                'dashboard': loaded_object
            }

        dashboard = loaded_object['dashboard']

        if loaded_object['version'] != self._dashboards_api_version:
            #
            # Convert the dashboard (if possible)
            #
            conversion_result, dashboard = convert_dashboard_between_versions(dashboard,
                                                                              loaded_object['version'],
                                                                              self._dashboards_api_version)

            if not conversion_result:
                return conversion_result, dashboard

        #
        # Create the new dashboard
        #
        return self.create_dashboard_from_template(dashboard_name, dashboard, filter, shared, public)

    def get_dashboard(self, dashboard_id):
        '''**Description**
            Return a dashboard with the pased in ID. This includes the dashboards created by the user and the ones shared with them by other users.

        **Success Return Value**
            A dictionary containing the requested dashboard data.

        **Example**
            `examples/dashboard_basic_crud.py <https://github.com/draios/python-sdc-client/blob/master/examples/dashboard_basic_crud.py>`_
        '''
        res = self.http.get(self.url + self._dashboards_api_endpoint + "/" + str(dashboard_id), headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def create_dashboard_from_dashboard(self, newdashname, templatename, filter=None, shared=False, public=False):
        '''**Description**
            Create a new dasboard using one of the existing dashboards as a template. You will be able to define the scope of the new dasboard.

        **Arguments**
            - **newdashname**: the name of the dashboard that will be created.
            - **viewname**: the name of the dasboard to use as the template, as it appears in the Sysdig Monitor dashboard page.
            - **filter**: a boolean expression combining Sysdig Monitor segmentation criteria defines what the new dasboard will be applied to. For example: *kubernetes.namespace.name='production' and container.image='nginx'*.
            - **shared**: if set to True, the new dashboard will be a shared one.
            - **public**: if set to True, the new dashboard will be shared with public token.

        **Success Return Value**
            A dictionary showing the details of the new dashboard.

        **Example**
            `examples/create_dashboard.py <https://github.com/draios/python-sdc-client/blob/master/examples/create_dashboard.py>`_
        '''
        #
        # Get the list of dashboards from the server
        #
        dashboard = self.http.get(self.url + self._dashboards_api_endpoint, params={"light": "true"}, headers=self.hdrs,
                                  verify=self.ssl_verify)
        if not self._checkResponse(dashboard):
            return [False, self.lasterr]

        j = dashboard.json()

        #
        # Find our template dashboard
        #
        dboard = None

        for db in j['dashboards']:
            if db['name'] == templatename:
                dboard = db
                break

        if dboard is None:
            self.lasterr = 'can\'t find dashboard ' + templatename + ' to use as a template'
            return [False, self.lasterr]

        ok, dboard = self.get_dashboard(dboard["id"])
        if not ok:
            return ok, dboard
        #
        # Create the dashboard
        #
        return self.create_dashboard_from_template(newdashname, dboard["dashboard"], filter, shared, public)

    def favorite_dashboard(self, dashboard_id, favorite):
        data = {"dashboard": {"favorite": favorite}}
        res = self.http.patch(self.url + self._dashboards_api_endpoint + "/" + str(dashboard_id), json=data,
                              headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def share_dashboard_with_all_teams(self, dashboard, mode="r"):
        role = "ROLE_RESOURCE_READ" if mode == "r" else "ROLE_RESOURCE_EDIT"
        dboard = copy.deepcopy(dashboard)
        dboard["sharingSettings"] = [
            {
                "member": {
                    "type": "USER_TEAMS",
                },
                "role": role,
            }
        ]
        dboard["shared"] = True

        return self.update_dashboard(dboard)

    def unshare_dashboard(self, dashboard):
        dboard = copy.deepcopy(dashboard)
        dboard["sharingSettings"] = []
        dboard["shared"] = False

        return self.update_dashboard(dboard)

    def share_dashboard_with_team(self, dashboard, team_id, mode="r"):
        role = "ROLE_RESOURCE_READ" if mode == "r" else "ROLE_RESOURCE_EDIT"
        dboard = copy.deepcopy(dashboard)

        if dboard["sharingSettings"] is None:
            dboard["sharingSettings"] = []

        dboard["sharingSettings"].append({
            "member": {
                "type": "TEAM",
                "id": team_id,
            },
            "role": role,
        })
        dboard["shared"] = True

        return self.update_dashboard(dboard)

    def create_dashboard_from_view(self, newdashname, viewname, filter, shared=False, public=False):
        '''**Description**
            Create a new dasboard using one of the Sysdig Monitor views as a template. You will be able to define the scope of the new dashboard.

        **Arguments**
            - **newdashname**: the name of the dashboard that will be created.
            - **viewname**: the name of the view to use as the template for the new dashboard. This corresponds to the name that the view has in the Explore page.
            - **filter**: a boolean expression combining Sysdig Monitor segmentation criteria that defines what the new dasboard will be applied to. For example: *kubernetes.namespace.name='production' and container.image='nginx'*.
            - **shared**: if set to True, the new dashboard will be a shared one.
            - **public**: if set to True, the new dashboard will be shared with public token.

        **Success Return Value**
            A dictionary showing the details of the new dashboard.

        **Example**
            `examples/create_dashboard.py <https://github.com/draios/python-sdc-client/blob/master/examples/create_dashboard.py>`_
        '''
        #
        # Find our template view
        #
        gvres = self.get_view(viewname)
        if gvres[0] is False:
            return gvres

        view = gvres[1]['dashboard']

        view['timeMode'] = {'mode': 1}
        view['time'] = {'last': 2 * 60 * 60 * 1000000, 'sampling': 2 * 60 * 60 * 1000000}

        #
        # Create the new dashboard
        #
        return self.create_dashboard_from_template(newdashname, view, filter, shared, public)

    def save_dashboard_to_file(self, dashboard, filename):
        '''
        **Description**
            Save a dashboard to disk. See :func:`~SdcClient.create_dashboard_from_file` to use the file to create a dashboard (usefl to create and restore backups).

            The file will contain a JSON object with the following properties:
            * version: dashboards API version (e.g. 'v2')
            * dashboard: dashboard object in the format of an array element returned by :func:`~SdcClient.get_dashboards`

        **Arguments**
            - **dashboard**: dashboard object in the format of an array element returned by :func:`~SdcClient.get_dashboards`
            - **filename**: name of a file that will contain a JSON object

        **Example**
            `examples/dashboard_save_load.py <https://github.com/draios/python-sdc-client/blob/master/examples/dashboard_save_load.py>`_
        '''
        with open(filename, 'w') as outf:
            json.dump({
                'version': self._dashboards_api_version,
                'dashboard': dashboard
            }, outf)

    def delete_dashboard(self, dashboard):
        '''**Description**
            Deletes a dashboard.

        **Arguments**
            - **dashboard**: the dashboard object as returned by :func:`~SdcClient.get_dashboards`.

        **Success Return Value**
            `None`.

        **Example**
            `examples/delete_dashboard.py <https://github.com/draios/python-sdc-client/blob/master/examples/delete_dashboard.py>`_
        '''
        if 'id' not in dashboard:
            return [False, "Invalid dashboard format"]

        res = self.http.delete(self.url + self._dashboards_api_endpoint + '/' + str(dashboard['id']), headers=self.hdrs,
                               verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, None]



import tatsu


def convert_scope_string_to_expression(scope=None):
    if scope is None or not scope:
        return [True, []]

    _SCOPE_GRAMMAR = """
        @@grammar::CALC

        start = expression $ ;

        expression
            =
            |   operand simple_operator word
            |   operand multiple_operator multiple_value
            ;

        simple_operator
            =
            |  'is not'
            |  'is'
            |  'contains'
            |  'does not contain'
            |  'starts with'
            |  '='
            ;

        multiple_operator
            =
            |  'not in'
            |  'in'
            ;

        operand = /[a-zA-Z0-9_\-\.]+/ ;

        multiple_value
            =
            | '[' word_array ']'
            | word
            ;

        word_array
            =
            | word ',' word_array
            | word
            ;

        word =
            | /[a-zA-Z0-9-_\-\.]+/
            | '"' /[a-zA-Z0-9-_\-\.]+/ '"'
            | "'" /[a-zA-Z0-9-_\-\.]+/ "'"
            ;
    """

    def flatten(S):
        if S == [] or S == ():
            return list(S)
        if isinstance(S[0], list) or isinstance(S[0], tuple):
            return flatten(S[0]) + flatten(S[1:])
        return list(S[:1]) + flatten(S[1:])

    try:
        grammar = tatsu.compile(_SCOPE_GRAMMAR)
        scope_list = []

        scope_expressions = scope.strip(' \t\n\r').split(' and ')

        for scope in scope_expressions:

            operand, parsed_operator, value = grammar.parse(scope)

            operator_match = {
                "is": "equals",
                "=": "equals",
                "is not": "notEquals",
                "in": "in",
                "not in": "notIn",
                "contains": "contains",
                "does not contain": "notContains",
                "starts with": "startsWith",
            }

            if isinstance(value, tuple) or isinstance(value, list):
                value = flatten(value)
                if len(value) > 1:
                    value = list(value[1:-1])  # Remove '[' and ']'
                    value = [elem for elem in value if elem != ',']  # Remove ','
            else:
                value = [value]

            operator = "" if parsed_operator not in operator_match else operator_match[parsed_operator]

            scope_list.append({
                'displayName': "",
                "isVariable": False,
                'operand': operand,
                'operator': operator,
                'value': value
            })
        return [True, scope_list]
    except Exception as ex:
        return [False, f"invalid scope: {scope}, {ex.message}"]



from ._activity_audit_v1 import ActivityAuditClientV1, ActivityAuditDataSource
from ._falco_rules_files_old import FalcoRulesFilesClientOld
from ._policy_events_old import PolicyEventsClientOld
from ._policy_events_v1 import PolicyEventsClientV1
from ._policy_v2 import (policy_action_capture, policy_action_kill, policy_action_pause, policy_action_stop,
                         PolicyClientV2)
from .scanning._alerts import ScanningAlertsClientV1

__all__ = ["PolicyEventsClientOld", "PolicyEventsClientV1", "FalcoRulesFilesClientOld",
           "PolicyClientV2", "policy_action_pause", "policy_action_stop", "policy_action_kill", "policy_action_capture",
           "ActivityAuditClientV1", "ActivityAuditDataSource", "ScanningAlertsClientV1"]



import json

from sdcclient._common import _SdcCommon


def policy_action_stop():
    return {
        "type": "POLICY_ACTION_STOP",
    }


def policy_action_capture(file_name, secs_before=5, secs_after=15, filter=""):
    return {
        "afterEventNs": secs_after * 1_000_000_000,
        "beforeEventNs": secs_before * 1_000_000_000,
        "isLimitedToContainer": False,
        "type": "POLICY_ACTION_CAPTURE",
        "filter": filter,
        "name": file_name,
        "bucketName": "",
        "storageType": "S3"
    }


def policy_action_pause():
    return {
        "type": "POLICY_ACTION_PAUSE",
    }


def policy_action_kill():
    return {
        "type": "POLICY_ACTION_KILL",
    }


class PolicyClientV2(_SdcCommon):
    def __init__(self, token="", sdc_url='https://secure.sysdig.com', ssl_verify=True, custom_headers=None):
        super(PolicyClientV2, self).__init__(token, sdc_url, ssl_verify, custom_headers)
        self.product = "SDS"

    def create_default_policies(self):
        """
        Create new policies based on the currently available set of rules. For now, this only covers Falco rules, but we might extend
        the endpoint later. The backend should use the defaultPolicies property of a previously provided FalcoRulesFiles model as
        guidance on the set of policies to create. The backend should only create new policies (not delete or modify), and should only
        create new policies if there is not an existing policy with the same name.

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the second the error or the result.

        Examples:
            >>> from sdcclient.secure import PolicyClientV2
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> ok, res = client.create_default_policies()
        """

        res = self.http.post(self.url + '/api/v2/policies/default', headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def delete_all_policies(self):
        """
        Delete all existing policies.

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the second the error or the result.

        Examples:
            >>> from sdcclient.secure import PolicyClientV2
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> ok, res = client.delete_all_policies()
        """

        ok, res = self.list_policies()
        if not ok:
            return False, res

        for policy in res:
            ok, res = self.delete_policy_id(policy["id"])
            if not ok:
                return False, res

        return True, "Policies Deleted"

    def list_policies(self):
        """
        List the current set of policies.

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the second the error or the result.

        Examples:
            >>> from sdcclient.secure import PolicyClientV2
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> ok, res = client.list_policies()
        """

        res = self.http.get(self.url + '/api/v2/policies', headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def get_policy(self, name):
        """
        Find the policy with name <name> and return its json description.

        Args:
            name(str): The name of the policy to fetch

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the
        second the error or the JSON object containing the policy.

        Examples:
            >>> import json
            >>> from sdcclient.secure import PolicyClientV2
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> ok, res = client.get_policy(name="Terminal shell in container")
            >>> if ok:
            >>>     print((json.dumps(res, indent=2)))
        """

        ok, res = self.list_policies()
        if not ok:
            return [False, res]
        policies = res

        # Find the policy with the given name and return it.
        for policy in policies:
            if policy["name"] == name:
                return [True, policy]

        return [False, "No policy with name {}".format(name)]

    def get_policy_id(self, id):
        """
        Find the policy with id <id> and return its json description.

        Args:
            id(int): The id of the policy to fetch

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the
        second the error or the JSON object containing the policy.

        Examples:
            >>> import json
            >>> from sdcclient.secure import PolicyClientV2
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> ok, res = client.get_policy_id(id=123456)
            >>> if ok:
            >>>     print((json.dumps(res, indent=2)))
        """

        res = self.http.get(self.url + '/api/v2/policies/{}'.format(id), headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def add_policy(self, name, description, rule_names=[], actions=[], scope=None, severity=0, enabled=True,
                   notification_channels=[], type="falco"):
        """
        Adds a new policy.

        Args:
            name(str): A short name for the policy
            description(str): Description of policy
            rule_names(list): Array of rule names. (They must be names instead of ids, as the rules list view is by name, to account for multiple rules having the same name).
            actions(list): It can be a `policy_action_stop()`, `policy_action_pause()`, `policy_action_capture()` or `policy_action_kill()` action
            scope(str): Where the policy is being applied- Container, Host etc.. (example: "container.image.repository = sysdig/agent")
            severity(int): True if the policy should be considered
            enabled(bool): How severe is this policy when violated. Range from 0 to 7 included.
            notification_channels(list): ids of the notification channels to subscribe to the policy
            type(str): Type of the Policy. It can be one of: `falco`, `list_matching`, `k8s_audit`.

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the second the error or the result.

        Examples:
            >>> from sdcclient.secure import PolicyClientV2, policy_action_stop
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> ok, res = client.add_policy(name="Terminal shell in container",
                                            description="A shell was spawned by a program in a container with an attached terminal.",
                                            rule_names=["Terminal shell in container"],
                                            actions=[policy_action_stop()],
                                            type="falco")
        """

        policy = {
            "name": name,
            "description": description,
            "ruleNames": rule_names,
            "actions": actions,
            "scope": scope,
            "severity": severity,
            "enabled": enabled,
            "notificationChannelIds": notification_channels,
            "type": type,
        }

        res = self.http.post(self.url + '/api/v2/policies', headers=self.hdrs, data=json.dumps(policy),
                             verify=self.ssl_verify)
        return self._request_result(res)

    def add_policy_json(self, policy_json):
        """
        Add a new policy using the provided json.

        Args:
            policy_json: a description of the new policy

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the second the error or the result.

        Examples:
            >>> import sys
            >>> from sdcclient.secure import PolicyClientV2
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> policy_json = sys.stdin.read()
            >>> ok, res = client.add_policy_json(policy_json)

        """

        try:
            policy_obj = json.loads(policy_json)
            if "origin" in policy_obj:
                del policy_obj["origin"]
        except Exception as e:
            return [False, "policy json is not valid json: {}".format(str(e))]

        res = self.http.post(self.url + '/api/v2/policies', headers=self.hdrs, data=json.dumps(policy_obj),
                             verify=self.ssl_verify)
        return self._request_result(res)

    def update_policy(self, id, name=None, description=None, rule_names=None, actions=None, scope=None,
                      severity=None, enabled=None, notification_channels=None):
        """
        Update policy with the provided values. Only the defined values will be updated.

        Args:
            id(int): The id of the policy to update
            name(str): A short name for the policy
            description(str): Description of policy
            rule_names(list): Array of rule names. (They must be names instead of ids, as the rules list view is by name, to account for multiple rules having the same name).
            actions(list): It can be a `policy_action_stop()`, `policy_action_pause()`, `policy_action_capture()` or `policy_action_kill()` action
            scope(str): Where the policy is being applied- Container, Host etc.. (example: "container.image.repository = sysdig/agent")
            severity(int): True if the policy should be considered
            enabled(bool): How severe is this policy when violated. Range from 0 to 7 included.
            notification_channels(list): ids of the notification channels to subscribe to the policy

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the second the error or the result.

        Examples:
            >>> from sdcclient.secure import PolicyClientV2, policy_action_stop
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> ok, res = client.update_policy(name="Terminal shell in container",
                                               description="A shell was spawned by a program in a container with an attached terminal.",
                                               rule_names=["Terminal shell in container"],
                                               actions=[policy_action_stop()])
        """

        ok, res = self.get_policy_id(id)
        if not ok:
            return [False, res]
        policy = res

        if name is not None:
            policy["name"] = name
        if description is not None:
            policy["description"] = description
        if rule_names is not None:
            policy["ruleNames"] = rule_names
        if actions is not None:
            policy["actions"] = actions
        if scope is not None:
            policy["scope"] = scope
        if severity is not None:
            policy["severity"] = severity
        if enabled is not None:
            policy["enabled"] = enabled
        if notification_channels is not None:
            policy["notificationChannelIds"] = notification_channels

        res = self.http.put(self.url + '/api/v2/policies/{}'.format(id), headers=self.hdrs, data=json.dumps(policy),
                            verify=self.ssl_verify)
        return self._request_result(res)

    def update_policy_json(self, policy_json):
        """
        Update an existing policy using the provided json. The 'id' field from the policy is
        used to determine which policy to update.

        Args:
            policy_json(str): A description of the new policy

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the second the error or the result.

        Examples:
            >>> import sys
            >>> from sdcclient.secure import PolicyClientV2
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> policy_json = sys.stdin.read()
            >>> ok, res = client.update_policy_json(policy_json)
        """

        try:
            policy_obj = json.loads(policy_json)
            if "origin" in policy_obj:
                del policy_obj["origin"]
        except Exception as e:
            return [False, "policy json is not valid json: {}".format(str(e))]

        if "id" not in policy_obj:
            return [False, "Policy Json does not have an 'id' field"]

        res = self.http.put(self.url + '/api/v2/policies/{}'.format(policy_obj["id"]), headers=self.hdrs,
                            data=json.dumps(policy_obj), verify=self.ssl_verify)
        return self._request_result(res)

    def delete_policy_name(self, name):
        """
        Delete the policy with the given name.

        Args:
            name(str): The name of the policy to delete

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the second the error or the JSON object representing the now-deleted policy.

        Examples:
            >>> from sdcclient.secure import PolicyClientV2
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> ok, res = client.delete_policy_name(name="Terminal shell in container")
        """

        ok, res = self.list_policies()
        if not ok:
            return [False, res]

        # Find the policy with the given name and delete it
        for policy in res:
            if policy["name"] == name:
                return self.delete_policy_id(policy["id"])

        return [False, "No policy with name {}".format(name)]

    def delete_policy_id(self, id):
        """
        Delete the policy with the given name.

        Args:
            id(int): The id of the policy to delete

        Returns: A touple (bool, res/err) where the first element indicates if the API call was successful and the second the error or the JSON object representing the now-deleted policy.

        Examples:
            >>> from sdcclient.secure import PolicyClientV2
            >>> client = PolicyClientV2(sdc_url="https://secure.sysdig.com", token=SECURE_TOKEN)
            >>> ok, res = client.delete_policy_id(id=123456)
        """

        res = self.http.delete(self.url + '/api/v2/policies/{}'.format(id), headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)



import json

from sdcclient._common import _SdcCommon


class EventsClientV1(_SdcCommon):
    def __init__(self, token="", sdc_url='https://app.sysdigcloud.com', ssl_verify=True, custom_headers=None):
        super().__init__(token, sdc_url, ssl_verify, custom_headers)
        self.product = "SDC"

    def get_events(self, from_s=None, to_s=None, last_s=None):
        '''**Description**
            Returns the list of Sysdig Monitor events.

        **Arguments**
            - **name**: filter events by name. Default: None.
            - **category**: filter events by category. Default: ['alert', 'custom', 'docker', 'containerd', 'kubernetes'].
            - **direction**: orders the list of events. Valid values: "before", "after". Default: "before".
            - **status**: status of the event as list. Default: ['triggered', 'resolved', 'acknowledged', 'unacknowledged']
            - **limit**: max number of events to retrieve. Default: 100.
            - **pivot**: event id to use as pivot. Default: None.

        **Success Return Value**
            A dictionary containing the list of events.

        **Example**
            `examples/list_events.py <https://github.com/draios/python-sdc-client/blob/master/examples/list_events.py>`_
        '''

        options = {
            "from": from_s,
            "to": to_s,
            "last": last_s,
        }
        params = {k: v for k, v in options.items() if v is not None}
        res = self.http.get(self.url + '/api/events/', headers=self.hdrs, params=params, verify=self.ssl_verify)
        return self._request_result(res)

    def get_event(self, id):
        """
        Retrieve an event using the ID
        Args:
            id(str): ID of the event to retrieve

        Returns:
            A tuple where the first parameter indicates if the call was successful,
            and the second parameter holds either the error as string, or the event matching this ID.

        Examples:
            >>> from sdcclient.monitor import EventsClientV1
            >>> client = EventsClientV1(token=SECURE_TOKEN)
            >>> ok, res = client.get_event(id='2343214984')
            >>> if ok:
            >>>     print(res["event"])
        """
        url = f'{self.url}/api/events/{id}'
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def post_event(self, name, description=None, severity=None, event_filter=None, tags=None):
        '''**Description**
            Send an event to Sysdig Monitor. The events you post are available in the Events tab in the Sysdig Monitor UI and can be overlied to charts.

        **Arguments**
            - **name**: the name of the new event.
            - **description**: a longer description offering detailed information about the event.
            - **severity**: syslog style from 0 (high) to 7 (low).
            - **event_filter**: metadata, in Sysdig Monitor format, of nodes to associate with the event, e.g. ``host.hostName = 'ip-10-1-1-1' and container.name = 'foo'``.
            - **tags**: a list of key-value dictionaries that can be used to tag the event. Can be used for filtering/segmenting purposes in Sysdig Monitor.

        **Success Return Value**
            A dictionary describing the new event.

        **Examples**
            - `examples/post_event_simple.py <https://github.com/draios/python-sdc-client/blob/master/examples/post_event_simple.py>`_
            - `examples/post_event.py <https://github.com/draios/python-sdc-client/blob/master/examples/post_event.py>`_
        '''
        options = {
            'name': name,
            'description': description,
            'severity': severity,
            'filter': event_filter,
            'tags': tags
        }
        edata = {
            'event': {k: v for k, v in options.items() if v is not None}
        }
        res = self.http.post(self.url + '/api/events/', headers=self.hdrs, data=json.dumps(edata),
                             verify=self.ssl_verify)
        return self._request_result(res)

    def delete_event(self, event):
        '''**Description**
            Deletes an event.

        **Arguments**
            - **event**: the event object as returned by :func:`~SdcClient.get_events`.

        **Success Return Value**
            `None`.

        **Example**
            `examples/delete_event.py <https://github.com/draios/python-sdc-client/blob/master/examples/delete_event.py>`_
        '''
        if 'id' not in event:
            return [False, "Invalid event format"]

        res = self.http.delete(self.url + '/api/events/' + str(event['id']), headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]
        return [True, None]



import json
import os
import shutil

import yaml

from sdcclient._common import _SdcCommon


class FalcoRulesFilesClientOld(_SdcCommon):
    def __init__(self, token="", sdc_url='https://secure.sysdig.com', ssl_verify=True, custom_headers=None):
        super(FalcoRulesFilesClientOld, self).__init__(token, sdc_url, ssl_verify, custom_headers)
        self.product = "SDS"

    # TODO: Remove this one, deprecated
    def _get_falco_rules(self, kind):
        res = self.http.get(self.url + '/api/settings/falco/{}RulesFile'.format(kind), headers=self.hdrs,
                            verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]
        data = res.json()
        return [True, data]

    # TODO: Change this one to use newestDefaultRulesFiles endpoint
    def get_system_falco_rules(self):
        '''**Description**
            Get the system falco rules file in use for this customer. See the `Falco wiki <https://github.com/draios/falco/wiki/Falco-Rules>`_ for documentation on the falco rules format.

        **Arguments**
            - None

        **Success Return Value**
            The contents of the system falco rules file.

        **Example**
            `examples/get_secure_system_falco_rules.py <https://github.com/draios/python-sdc-client/blob/master/examples/get_secure_system_falco_rules.py>`_
        '''

        return self._get_falco_rules("system")

    def get_user_falco_rules(self):
        '''**Description**
            Get the user falco rules file in use for this customer. See the `Falco wiki <https://github.com/draios/falco/wiki/Falco-Rules>`_ for documentation on the falco rules format.

        **Arguments**
            - None

        **Success Return Value**
            The contents of the user falco rules file.

        **Example**
            `examples/get_secure_user_falco_rules.py <https://github.com/draios/python-sdc-client/blob/master/examples/newestDefaultRulesFiles>`_
        '''
        ok, res = self._get_user_falco_rules()
        if not ok:
            return [False, res]

        local_rules_file = [file
                            for file in res["customFalcoRulesFiles"]["files"]
                            if file["name"] == "falco_rules_local.yaml"]
        if len(local_rules_file) == 0:
            return [False, "Expected falco_rules_local.yaml file, but no file found"]

        return [True, local_rules_file[0]["variants"][0]["content"]]

    def _get_user_falco_rules(self):
        res = self.http.get(self.url + '/api/settings/falco/customRulesFiles', headers=self.hdrs,
                            verify=self.ssl_verify)

        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, (res.json())]

    # TODO: Remove this
    def _set_falco_rules(self, kind, rules_content):
        payload = self._get_falco_rules(kind)

        if not payload[0]:
            return payload

        payload[1]["{}RulesFile".format(kind)]["content"] = rules_content  # pylint: disable=unsubscriptable-object

        res = self.http.put(self.url + '/api/settings/falco/{}RulesFile'.format(kind), headers=self.hdrs,
                            data=json.dumps(payload[1]), verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]
        return [True, res.json()]

    def set_system_falco_rules(self, rules_content):
        '''**Description**
            Set the system falco rules file in use for this customer. NOTE: This API endpoint can *only* be used in on-premise deployments. Generally the system falco rules file is only modified in conjunction with Sysdig support. See the `Falco wiki <https://github.com/draios/falco/wiki/Falco-Rules>`_ for documentation on the falco rules format.

        **Arguments**
            - A string containing the system falco rules.

        **Success Return Value**
            The contents of the system falco rules file that were just updated.

        **Example**
            `examples/set_secure_system_falco_rules.py <https://github.com/draios/python-sdc-client/blob/master/examples/get_secure_system_falco_rules.py>`_

        '''
        return self._set_falco_rules("system", rules_content)

    def set_user_falco_rules(self, rules_content):
        '''**Description**
            Set the user falco rules file in use for this customer. See the `Falco wiki <https://github.com/draios/falco/wiki/Falco-Rules>`_ for documentation on the falco rules format.

        **Arguments**
            - A string containing the user falco rules.

        **Success Return Value**
            The contents of the user falco rules file that were just updated.

        **Example**
            `examples/set_secure_user_falco_rules.py <https://github.com/draios/python-sdc-client/blob/master/examples/get_secure_user_falco_rules.py>`_

        '''
        ok, res = self._get_user_falco_rules()

        if not ok:
            return res

        local_rules_file = [file
                            for file in res["customFalcoRulesFiles"]["files"]
                            if file["name"] == "falco_rules_local.yaml"]
        if len(local_rules_file) == 0:
            return [False, "Expected falco_rules_local.yaml file, but no file found"]

        local_rules_file[0]["variants"][0]["content"] = rules_content

        res = self.http.put(self.url + '/api/settings/falco/customRulesFiles', headers=self.hdrs,
                            data=json.dumps(res), verify=self.ssl_verify)

        if not self._checkResponse(res):
            return [False, self.lasterr]
        res_json = res.json()
        return [True, res_json["customFalcoRulesFiles"]["files"][0]["variants"][0]["content"]]

    # get_falco_syscall_rules()

    # get_falco_ka_rules()

    # Only one kind for now called "default", but might add a "custom" kind later.
    # TODO Remove this one
    def _get_falco_rules_files(self, kind):

        res = self.http.get(self.url + '/api/settings/falco/{}RulesFiles'.format(kind), headers=self.hdrs,
                            verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]
        data = res.json()

        return [True, data]

    def get_default_falco_rules_files(self):
        '''**Description**
            Get the set of falco rules files from the backend. The _files programs and endpoints are a
               replacement for the system_file endpoints and allow for publishing multiple files instead
               of a single file as well as publishing multiple variants of a given file that are compatible
               with different agent versions.

        **Arguments**
            - None

        **Success Return Value**
            A dict with the following keys:
               - tag: A string used to uniquely identify this set of rules. It is recommended that this tag change every time the set of rules is updated.
               - files: An array of dicts. Each dict has the following keys:
                     - name: the name of the file
                     - variants: An array of dicts with the following keys:
                          - requiredEngineVersion: the minimum falco engine version that can read this file
                          - content: the falco rules content
            An example would be:
               {'tag': 'v1.5.9',
                'files': [
                    {
                        'name': 'falco_rules.yaml',
                        'variants': [
                            {
                                'content': '- required_engine_version: 29\n\n- list: foo\n',
                                'requiredEngineVersion': 29
                            },
                            {
                                'content': '- required_engine_version: 1\n\n- list: foo\n',
                                'requiredEngineVersion': 1
                            }
                        ]
                    },
                    {
                        'name': 'k8s_audit_rules.yaml',
                        'variants': [
                            {
                                'content': '# some comment\n',
                                'requiredEngineVersion': 0
                            }
                        ]
                    }
                ]
              }

        **Example**
            `examples/get_default_falco_rules_files.py <https://github.com/draios/python-sdc-client/blob/master/examples/get_default_falco_rules_files.py>`_
        '''

        res = self._get_falco_rules_files("default")

        if not res[0]:
            return res
        else:
            res_obj = res[1]["defaultFalcoRulesFiles"]

            # Copy only the tag and files over
            ret = {}

            if "tag" in res_obj:
                ret["tag"] = res_obj["tag"]

            if "files" in res_obj:
                ret["files"] = res_obj["files"]

            if "defaultPolicies" in res_obj:
                ret["defaultPolicies"] = res_obj["defaultPolicies"]

            return [True, ret]

    def save_default_falco_rules_files(self, fsobj, save_dir):
        '''**Description**
            Given a dict returned from get_default_falco_rules_files, save those files to a set of files below save_dir.
               The first level below save_dir is a directory with the tag name and an optional default_policies.yaml file,
               which groups rules into recommended default policies. The second level is a directory per file.
               The third level is a directory per variant. Finally the files are at the lowest level, in a file called "content".
            For example, using the example dict in get_default_falco_rules_files(), the directory layout would look like:
                save_dir/
                    default_policies.yaml
                    v1.5.9/
                        falco_rules.yaml/
                            29/
                                content: a file containing "- required_engine_version: 29\n\n- list: foo\n"
                            1/
                                content: a file containing "- required_engine_version: 1\n\n- list: foo\n"
                        k8s_audit_rules.yaml/
                            0/
                                content: a file containing "# some comment"
        **Arguments**
            - fsobj: a python dict matching the structure returned by get_default_falco_rules_files()
            - save_dir: a directory path under which to save the files. If the path already exists, it will be removed first.

        **Success Return Value**
            - None

        **Example**
            `examples/get_default_falco_rules_files.py <https://github.com/draios/python-sdc-client/blob/master/examples/get_default_falco_rules_files.py>`_
        '''
        if os.path.exists(save_dir):
            try:
                if os.path.isdir(save_dir):
                    shutil.rmtree(save_dir)
                else:
                    os.unlink(save_dir)
            except Exception as e:
                return [False, "Could not remove existing save dir {}: {}".format(save_dir, str(e))]

        prefix = os.path.join(save_dir, fsobj["tag"])
        try:
            os.makedirs(prefix)
        except Exception as e:
            return [False, "Could not create tag directory {}: {}".format(prefix, str(e))]

        if "defaultPolicies" in fsobj:
            with open(os.path.join(save_dir, "default_policies.yaml"), 'w') as outfile:
                yaml.safe_dump(fsobj["defaultPolicies"], outfile)

        if "files" in fsobj:
            for fobj in fsobj["files"]:
                fprefix = os.path.join(prefix, fobj["name"])
                try:
                    os.makedirs(fprefix)
                except Exception as e:
                    return [False, "Could not create file directory {}: {}".format(fprefix, str(e))]
                for variant in fobj["variants"]:
                    vprefix = os.path.join(fprefix, str(variant["requiredEngineVersion"]))
                    try:
                        os.makedirs(vprefix)
                    except Exception as e:
                        return [False, "Could not create variant directory {}: {}".format(vprefix, str(e))]
                    cpath = os.path.join(vprefix, "content")
                    try:
                        with open(cpath, "w") as cfile:
                            cfile.write(variant["content"])
                    except Exception as e:
                        return [False, "Could not write content to {}: {}".format(cfile, str(e))]

        return [True, None]

    # Only One kind for now, but might add a "custom" kind later.
    def _set_falco_rules_files(self, kind, rules_files):

        payload = self._get_falco_rules_files(kind)

        if not payload[0]:
            return payload

        obj = payload[1]["{}FalcoRulesFiles".format(kind)]  # pylint: disable=unsubscriptable-object

        obj["tag"] = rules_files["tag"]
        obj["files"] = rules_files["files"]
        if "defaultPolicies" in rules_files:
            obj["defaultPolicies"] = rules_files["defaultPolicies"]

        res = self.http.put(self.url + '/api/settings/falco/{}RulesFiles'.format(kind), headers=self.hdrs,
                            data=json.dumps(payload[1]), verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]
        return [True, res.json()]

    def set_default_falco_rules_files(self, rules_files):
        '''**Description**
            Update the set of falco rules files to the provided set of files. See the `Falco wiki <https://github.com/draios/falco/wiki/Falco-Rules>`_ for documentation on the falco rules format.
               The _files programs and endpoints are a replacement for the system_file endpoints and
               allow for publishing multiple files instead of a single file as well as publishing
               multiple variants of a given file that are compatible with different agent versions.

        **Arguments**
            - rules_files: a dict with the same structure as returned by get_default_falco_rules_files.

        **Success Return Value**
            The contents of the default falco rules files that were just updated.

        **Example**
            `examples/set_default_falco_rules_files.py <https://github.com/draios/python-sdc-client/blob/master/examples/set_default_falco_rules_files.py>`_

        '''

        return self._set_falco_rules_files("default", rules_files)

    def load_default_falco_rules_files(self, save_dir):
        '''**Description**
            Given a file and directory layout as described in save_default_falco_rules_files(), load those files and
            return a dict representing the contents. This dict is suitable for passing to set_default_falco_rules_files().

        **Arguments**
            - save_dir: a directory path from which to load the files.

        **Success Return Value**
            - A dict matching the format described in get_default_falco_rules_files.

        **Example**
            `examples/set_default_falco_rules_files.py <https://github.com/draios/python-sdc-client/blob/master/examples/set_default_falco_rules_files.py>`_
        '''

        tags = os.listdir(save_dir)

        try:
            tags.remove("default_policies.yaml")
        except ValueError:
            # Do nothing, it wasn't in the list of files
            pass

        if len(tags) != 1:
            return [False, "Directory {} did not contain exactly 1 entry".format(save_dir)]

        tpath = os.path.join(save_dir, tags[0])

        if not os.path.isdir(tpath):
            return [False, "Tag path {} is not a directory".format(tpath)]

        defjson = []
        defpath = os.path.join(save_dir, "default_policies.yaml")
        if os.path.exists(defpath):
            try:
                with open(defpath, "r") as infile:
                    defjson = yaml.safe_load(infile)
            except Exception as exc:
                return [False, "Could not load default_policies.yaml: " + exc]

        ret = {"tag": os.path.basename(tpath), "files": [], "defaultPolicies": defjson}

        for fdir in os.listdir(tpath):
            fpath = os.path.join(tpath, fdir)
            if not os.path.isdir(fpath):
                return [False, "File path {} is not a directory".format(fpath)]
            fobj = {"name": os.path.basename(fpath), "variants": []}
            for vdir in os.listdir(fpath):
                vpath = os.path.join(fpath, vdir)
                if not os.path.isdir(vpath):
                    return [False, "Variant path {} is not a directory".format(vpath)]
                cpath = os.path.join(vpath, "content")
                try:
                    with open(cpath, 'r') as content_file:
                        try:
                            required_engine_version = int(os.path.basename(vpath))
                            if int(os.path.basename(vpath)) < 0:
                                return [False, "Variant directory {} must be a positive number".format(vpath)]
                            fobj["variants"].append({
                                "requiredEngineVersion": required_engine_version,
                                "content": content_file.read()
                            })
                        except ValueError:
                            return [False, "Variant directory {} must be a number".format(vpath)]
                except Exception as e:
                    return [False, "Could not read content at {}: {}".format(cpath, str(e))]

            ret["files"].append(fobj)

        return [True, ret]



import json

from sdcclient._common import _SdcCommon


class ScanningAlertsClientV1(_SdcCommon):
    def __init__(self, token="", sdc_url='https://secure.sysdig.com', ssl_verify=True, custom_headers=None):
        super(ScanningAlertsClientV1, self).__init__(token, sdc_url, ssl_verify, custom_headers)
        self.product = "SDS"

    class RepositoryAlertTrigger:
        @staticmethod
        def new_image_analyzed(alert):
            alert["triggers"]["analysis_update"] = True

        @staticmethod
        def scan_result_change_fail(alert):
            alert["triggers"]["policy_eval"] = True
            alert["onlyPassFail"] = True

        @staticmethod
        def scan_result_change_any(alert):
            alert["triggers"]["policy_eval"] = True
            alert["onlyPassFail"] = False

        @staticmethod
        def cve_update(alert):
            alert["triggers"]["vuln_update"] = True

    def add_repository_alert(self, name, registry, repository, tag, description="", triggers=None, notification_channels=None, enabled=True):
        '''
        Create a new repository alert

        Args:
            name(str): The name of the alert.
            registry(str): Registry to alert (e.g. docker.io)
            repository(str): Repository to alert (e.g. sysdig/agent)
            tag(str): Tag to alert (e.g. latest)
            description(str): The description of the alert.
            triggers(list): A list of RepositoryAlertTrigger indicating which triggers should be enabled. (default: [ScanningAlertsClientV1.RuntimeAlertTrigger.new_image_analyzed])
            notification_channels(list): A list of notification channel ids.
            enabled(bool): Whether this alert should actually be applied. Defaults to true.
        Returns:
            The created alert.

        Examples:
            >>> client = ScanningAlertsClientV1(sdc_url=os.getenv("SDC_SECURE_URL", "https://secure.sysdig.com"),
            >>>                                 token=os.getenv("SDC_SECURE_TOKEN"))
            >>> ok, res = client.add_repository_alert(
            >>>        name="A name",
            >>>        registry="docker.io",
            >>>        repository="sysdig/agent",
            >>>        tag="latest",
            >>>        description="A description",
            >>>        triggers=[ScanningAlertsClientV1.RepositoryAlertTrigger.new_image_analyzed,
            >>>                  ScanningAlertsClientV1.RepositoryAlertTrigger.scan_result_change_fail,
            >>>                  ScanningAlertsClientV1.RepositoryAlertTrigger.cve_update]
            >>>)
            >>> if not ok:
            >>>    print(f"error creating alert: {res}")
            >>> alert_id = res["alertId"]
        '''
        if not triggers:
            triggers = [ScanningAlertsClientV1.RepositoryAlertTrigger.new_image_analyzed]

        alert = {
                'name':                   name,
                'description':            description,
                'type':                   'repository',
                'triggers':               {
                        "unscanned":       False,
                        "analysis_update": False,
                        "vuln_update":     False,
                        "policy_eval":     False,
                        "failed":          False
                },
                'repositories':           [{
                        'registry':   registry,
                        'repository': repository,
                        'tag':        tag,
                }],
                "onlyPassFail":           False,
                "skipEventSend":          False,
                'enabled':                enabled,
                'notificationChannelIds': notification_channels,
        }

        for trigger in triggers:
            trigger(alert)

        res = self.http.post(f"{self.url}/api/scanning/v1/alerts", headers=self.hdrs, data=json.dumps(alert), verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def update_repository_alert(self, id, name=None, registry=None, repository=None, tag=None, description=None, triggers=None, notification_channels=None, enabled=None):
        '''
        Updates a repository alert. Fields that are not specified, will not be modified.

        Args:
            id(str): Alert ID.
            name(str): The name of the alert.
            registry(str): Registry to alert (e.g. docker.io)
            repository(str): Repository to alert (e.g. sysdig/agent)
            tag(str): Tag to alert (e.g. latest)
            description(str): The description of the alert.
            triggers(list): A list of RepositoryAlertTrigger indicating which triggers should be enabled. (default: [ScanningAlertsClientV1.RuntimeAlertTrigger.unscanned_image])
            notification_channels(list): A list of notification channel ids.
            enabled(bool): Whether this alert should actually be applied. Defaults to true.
        Returns:
            The updated alert.

        Examples:
            >>> client = ScanningAlertsClientV1(sdc_url=os.getenv("SDC_SECURE_URL", "https://secure.sysdig.com"),
            >>>                                 token=os.getenv("SDC_SECURE_TOKEN"))
            >>> ok, res = client.update_repository_alert(
            >>>     id=alert_id,
            >>>     name="An updated name",
            >>>     registry="updated_registry",
            >>>     repository="updated_repository",
            >>>     tag="v1",
            >>>     description="An updated description",
            >>>     triggers=[ScanningAlertsClientV1.RepositoryAlertTrigger.scan_result_change_fail]
            >>> )
            >>> if not ok:
            >>>    print(f"error updating alert: {res}")
            >>> alert_id = res["alertId"]
        '''
        ok, alert = self.get_alert(id)
        if not ok:
            return False, f"unable to retrieve alert by ID {id}: {alert}"

        if name is not None:
            alert["name"] = name
        if description is not None:
            alert["description"] = description
        if registry is not None:
            alert["repositories"][0]["registry"] = registry
        if repository is not None:
            alert["repositories"][0]["repository"] = repository
        if tag is not None:
            alert["repositories"][0]["tag"] = tag
        if triggers is not None:
            alert["triggers"] = {
                    "unscanned":       False,
                    "analysis_update": False,
                    "vuln_update":     False,
                    "policy_eval":     False,
                    "failed":          False
            }
            alert["onlyPassFail"] = False
            for trigger in triggers:
                trigger(alert)
        if notification_channels is not None:
            alert["notificationChannelIds"] = notification_channels
        if enabled is not None:
            alert["enabled"] = enabled

        res = self.http.put(f"{self.url}/api/scanning/v1/alerts/{id}", headers=self.hdrs, data=json.dumps(alert), verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    class RuntimeAlertTrigger:
        @staticmethod
        def unscanned_image(alert):
            alert["triggers"]["unscanned"] = True

        @staticmethod
        def scan_result_change_fail(alert):
            alert["triggers"]["policy_eval"] = True
            alert["onlyPassFail"] = True

        @staticmethod
        def scan_result_change_any(alert):
            alert["triggers"]["policy_eval"] = True
            alert["onlyPassFail"] = False

        @staticmethod
        def cve_update(alert):
            alert["triggers"]["vuln_update"] = True

    def add_runtime_alert(self, name, description="", scope="", triggers=None, notification_channels=None, enabled=True):
        '''
        Create a new runtime alert

        Args:
            name(str): The name of the alert.
            description(str): The description of the alert.
            scope(str): An AND-composed string of predicates that selects the scope in which the alert will be applied. (like: 'host.domain = "example.com" and container.image != "alpine:latest"')
            triggers(list): A list of RuntimeAlertTrigger indicating which triggers should be enabled. (default: [ScanningAlertsClientV1.RuntimeAlertTrigger.unscanned_image])
            notification_channels(list): A list of notification channel ids.
            enabled(bool): Whether this alert should actually be applied. Defaults to true.
        Returns:
            The created alert.

        Examples:
            >>> client = ScanningAlertsClientV1(sdc_url=os.getenv("SDC_SECURE_URL", "https://secure.sysdig.com"),
            >>>                                 token=os.getenv("SDC_SECURE_TOKEN"))
            >>> ok, res = client.add_runtime_alert(
            >>>     name="A name",
            >>>     description="A description",
            >>>     scope="",
            >>>     triggers=[ScanningAlertsClientV1.RuntimeAlertTrigger.unscanned_image,
            >>>               ScanningAlertsClientV1.RuntimeAlertTrigger.scan_result_change_fail,
            >>>               ScanningAlertsClientV1.RuntimeAlertTrigger.cve_update]
            >>>)
            >>> if not ok:
            >>>    print(f"error creating alert: {res}")
            >>> alert_id = res["alertId"]
        '''
        if not triggers:
            triggers = [ScanningAlertsClientV1.RuntimeAlertTrigger.unscanned_image]

        alert = {
                'name':                   name,
                'description':            description,
                'type':                   'runtime',
                'triggers':               {
                        "unscanned":       False,
                        "analysis_update": False,
                        "vuln_update":     False,
                        "policy_eval":     False,
                        "failed":          False
                },
                'scope':                  scope,
                "onlyPassFail":           False,
                "skipEventSend":          False,
                'enabled':                enabled,
                'notificationChannelIds': notification_channels,
        }

        for trigger in triggers:
            trigger(alert)

        res = self.http.post(f"{self.url}/api/scanning/v1/alerts", headers=self.hdrs, data=json.dumps(alert), verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def update_runtime_alert(self, id, name=None, description=None, scope=None, triggers=None, notification_channels=None, enabled=None):
        '''
        Updates a runtime alert. Fields that are not specified, will not be modified.

        Args:
            id(str): Alert ID.
            name(str): The name of the alert.
            description(str): The description of the alert.
            scope(str): An AND-composed string of predicates that selects the scope in which the alert will be applied. (like: 'host.domain = "example.com" and container.image != "alpine:latest"')
            triggers(list): A list of RuntimeAlertTrigger indicating which triggers should be enabled. (default: [ScanningAlertsClientV1.RuntimeAlertTrigger.unscanned_image])
            notification_channels(list): A list of notification channel ids.
            enabled(bool): Whether this alert should actually be applied. Defaults to true.
        Returns:
            The updated alert.

        Examples:
            >>> client = ScanningAlertsClientV1(sdc_url=os.getenv("SDC_SECURE_URL", "https://secure.sysdig.com"),
            >>>                                 token=os.getenv("SDC_SECURE_TOKEN"))
            >>> ok, res = client.update_runtime_alert(
            >>>     id=alert_id,
            >>>     name="An updated name",
            >>>     description="An updated description",
            >>>     scope="agent.id = 'foo'",
            >>>     triggers=[ScanningAlertsClientV1.RuntimeAlertTrigger.scan_result_change_fail]
            >>> )
            >>> if not ok:
            >>>    print(f"error updating alert: {res}")
            >>> alert_id = res["alertId"]
        '''
        ok, alert = self.get_alert(id)
        if not ok:
            return False, f"unable to retrieve alert by ID {id}: {alert}"

        if name is not None:
            alert["name"] = name
        if description is not None:
            alert["description"] = description
        if scope is not None:
            alert["scope"] = scope
        if triggers is not None:
            alert["triggers"] = {
                    "unscanned":       False,
                    "analysis_update": False,
                    "vuln_update":     False,
                    "policy_eval":     False,
                    "failed":          False
            }
            alert["onlyPassFail"] = False
            for trigger in triggers:
                trigger(alert)
        if notification_channels is not None:
            alert["notificationChannelIds"] = notification_channels
        if enabled is not None:
            alert["enabled"] = enabled

        res = self.http.put(f"{self.url}/api/scanning/v1/alerts/{id}", headers=self.hdrs, data=json.dumps(alert), verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def get_alert(self, alertid):
        '''
        Retrieve the scanning alert with the given id

        Args:
            alertid: Unique identifier associated with this alert.

        Returns:
            A JSON object containing the alert description.

        Examples:
            >>> client = ScanningAlertsClientV1(sdc_url=os.getenv("SDC_SECURE_URL", "https://secure.sysdig.com"),
            >>>                                 token=os.getenv("SDC_SECURE_TOKEN"))
            >>> ok, res = client.get_alert(alert_id)
            >>> if not ok:
            >>>     print(f"error retrieving alert {alert_id}: {res}")
            >>> alert = res
        '''

        res = self.http.get(f"{self.url}/api/scanning/v1/alerts/{alertid}", headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]
        return [True, res.json()]

    def list_alerts(self, limit=None, cursor=None):
        '''
        List the current set of scanning alerts.
        Args:
            limit(int): Maximum number of alerts in the response.
            cursor: An opaque string representing the current position in the list of alerts. It's provided in the 'responseMetadata' of the list_alerts response.

        Returns:
            A JSON containing the list of alerts in the 'alerts' field, and the current cursor position in the 'responseMetadata' field.

        Examples:
            >>> client = ScanningAlertsClientV1(sdc_url=os.getenv("SDC_SECURE_URL", "https://secure.sysdig.com"),
            >>>                                 token=os.getenv("SDC_SECURE_TOKEN"))
            >>> ok, res =client.list_alerts()
            >>> if not ok:
            >>>     print(f"error listing alerts: {res}")
            >>> for alert in res["alerts"]:
            >>>     print(alert["alertId"])
            >>>
            >>> # Load more alerts
            >>> if res["responseMetadata"] is not None:
            >>>     ok, res = client.list_alerts(cursor=res["responseMetadata"]["next_cursor"])
        '''

        url = f"{self.url}/api/scanning/v1/alerts"
        if limit:
            url += '?limit=' + str(limit)
            if cursor:
                url += '&cursor=' + cursor

        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def delete_alert(self, policyid):  # FIXME: policyid must be maintained for backwards compatibility reasons with older versions, but should be renamed to id or alert_id
        '''
        Delete the alert with the given id

        Args:
            policyid:  Unique identifier associated with this alert.

        Examples:
            >>> client = ScanningAlertsClientV1(sdc_url=os.getenv("SDC_SECURE_URL", "https://secure.sysdig.com"),
            >>>                                 token=os.getenv("SDC_SECURE_TOKEN"))
            >>> client.delete_alert(alert_id)
        '''

        res = self.http.delete(f"{self.url}/api/scanning/v1/alerts/{policyid}", headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]
        return [True, res.text]

    def add_alert_object(self, object):
        '''
        Adds alert object as raw JSON object.

        Args:
            object: JSON repsentation of the alert.

        Examples:
            >>> client = ScanningAlertsClientV1(sdc_url=os.getenv("SDC_SECURE_URL", "https://secure.sysdig.com"),
            >>>                                 token=os.getenv("SDC_SECURE_TOKEN"))
            >>> alert = {
            >>>     "enabled": True,
            >>>     "type": "runtime",
            >>>     "name": "runtime-scanning-alert",
            >>>     "triggers": {
            >>>         "unscanned": True,
            >>>         "analysis_update": False,
            >>>         "vuln_update": False,
            >>>         "policy_eval": False,
            >>>         "failed": False
            >>>     },
            >>>     "autoscan": False,
            >>>     "onlyPassFail": False,
            >>>     "skipEventSend": False,
            >>>     "notificationChannelIds": []
            >>> }
            >>> client.add_alert_object(alert)
        '''
        url = self.url + '/api/scanning/v1/alerts'
        data = json.dumps(object)
        res = self.http.post(url, headers=self.hdrs, data=data, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]



import json
import time

from sdcclient._common import _SdcCommon
from sdcclient.secure import FalcoRulesFilesClientOld, PolicyEventsClientV1, PolicyEventsClientOld, PolicyClientV2


class SdSecureClient(FalcoRulesFilesClientOld,
                     PolicyEventsClientV1,
                     PolicyEventsClientOld,
                     PolicyClientV2,
                     _SdcCommon):
    def __init__(self, token="", sdc_url='https://secure.sysdig.com', ssl_verify=True, custom_headers=None):
        super(SdSecureClient, self).__init__(token, sdc_url, ssl_verify, custom_headers)

        self.product = "SDS"
        self._policy_v2 = None

    @property
    def policy_v2(self):
        '''**Description**
            True if policy V2 API is available
        '''
        if self._policy_v2 is None:
            res = self.http.get(self.url + '/api/v2/policies/default', headers=self.hdrs, verify=self.ssl_verify)
            self._policy_v2 = res.status_code != 404
        return self._policy_v2

    def list_rules(self):
        '''**Description**
            Returns the list of rules in the system. These are grouped by name
            and do not necessarily represent individual rule objects, as multiple
            rules can have the same name.

        **Arguments**
            - None

        **Success Return Value**
            A JSON object representing the list of rules.
        '''
        res = self.http.get(self.url + '/api/secure/rules/summaries', headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def get_rules_group(self, name):
        '''**Description**
            Retrieve a group of all rules having the given name. This is used to
            show how a base rule is modified by later rules that override/append
            to the rule.

        **Arguments**
            - name: the name of the rule group

        **Success Return Value**
            A JSON object representing the list of rules.
        '''
        res = self.http.get(self.url + '/api/secure/rules/groups?name={}'.format(name), headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def get_rule_id(self, id):
        '''**Description**
            Retrieve info about a single rule

        **Arguments**
            - id: the id of the rule

        **Success Return Value**
            A JSON object representing the rule.
        '''
        res = self.http.get(self.url + '/api/secure/rules/{}'.format(id), headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def add_rule(self, name, details={}, description="", tags=[]):
        '''**Description**
            Create a new rule

        **Arguments**
            - name: A name for this object. Should exactly be the value of the "rule" property of the yaml object.
            - details: The rule description as a python dictionary.
            - description: A description of this rule. No newlines/formatting.
            - tags: The set of tags.

        **Success Return Value**
            A JSON object representing the rule.
        '''
        rule = {
            "name": name,
            "description": description,
            "details": details,
            "tags": tags
        }
        res = self.http.post(self.url + '/api/secure/rules', data=json.dumps(rule), headers=self.hdrs,
                             verify=self.ssl_verify)
        return self._request_result(res)

    def update_rule(self, id, details={}, description="", tags=[]):
        '''**Description**
            Update info associated with a rule

        **Arguments**
            - id: The rule id
            - details: The rule description as a python dictionary.
            - description: A description of this rule. No newlines/formatting.
            - tags: The set of tags.

        **Success Return Value**
            A JSON object representing the rule.
        '''
        ok, res = self.get_rule_id(id)
        if not ok:
            return [False, res]
        rule = res

        if details:
            rule['details'] = details
        if description:
            rule['description'] = description
        if tags:
            rule['tags'] = tags
        res = self.http.put(self.url + '/api/secure/rules/{}'.format(id), data=json.dumps(rule), headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def delete_rule(self, id):
        '''**Description**
            Delete the rule with given id.

        **Arguments**
            - id: The rule id

        **Success Return Value**
            A JSON object representing the rule.
        '''
        res = self.http.delete(self.url + '/api/secure/rules/{}'.format(id), headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def list_falco_macros(self):
        '''**Description**
            Returns the list of macros in the system. These are grouped by name
            and do not necessarily represent individual macro objects, as multiple
            macros can have the same name.

        **Arguments**
            - None

        **Success Return Value**
            A JSON object representing the list of falco macros.
        '''
        res = self.http.get(self.url + '/api/secure/falco/macros/summaries', headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def get_falco_macros_group(self, name):
        '''**Description**
            Retrieve a group of all falco groups having the given name. This is used
            to show how a base macro is modified by later macrosthat override/append
            to the macro.

        **Arguments**
            - name: the name of the falco macros group

        **Success Return Value**
            A JSON object representing the list of falco macros.
        '''
        res = self.http.get(self.url + '/api/secure/falco/macros/groups?name={}'.format(name), headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def get_falco_macro_id(self, id):
        '''**Description**
            Retrieve info about a single falco macro

        **Arguments**
            - id: the id of the falco macro

        **Success Return Value**
            A JSON object representing the falco macro.
        '''
        res = self.http.get(self.url + '/api/secure/falco/macros/{}'.format(id), headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def add_falco_macro(self, name, condition, append=False):
        '''**Description**
            Create a new macro

        **Arguments**
            - name: A name for this object. Should exactly be the value of the "macro" property of the yaml object.
            - condition: the full condition text exactly as represented in the yaml file.

        **Success Return Value**
            A JSON object representing the falco macro.
        '''
        macro = {
            "name": name,
            "condition": {
                "components": [],
                "condition": condition
            },
            "append": append
        }
        res = self.http.post(self.url + '/api/secure/falco/macros', data=json.dumps(macro), headers=self.hdrs,
                             verify=self.ssl_verify)
        return self._request_result(res)

    def update_falco_macro(self, id, condition):
        '''**Description**
            Update info associated with a macro

        **Arguments**
            - id: The rule id
            - condition: the full condition text exactly as represented in the yaml file.

        **Success Return Value**
            A JSON object representing the macro.
        '''
        ok, res = self.get_falco_macro_id(id)
        if not ok:
            return [False, res]
        macro = res
        macro['condition']['condition'] = condition

        res = self.http.put(self.url + '/api/secure/falco/macros/{}'.format(id), data=json.dumps(macro),
                            headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def delete_falco_macro(self, id):
        '''**Description**
            Delete the macro with given id.

        **Arguments**
            - id: The macro id

        **Success Return Value**
            A JSON object representing the macro.
        '''
        res = self.http.delete(self.url + '/api/secure/falco/macros/{}'.format(id), headers=self.hdrs,
                               verify=self.ssl_verify)
        return self._request_result(res)

    def list_falco_lists(self):
        '''**Description**
            Returns the list of falco lists in the system. These are grouped by
            name and do not necessarily represent individual falco list objects,
            as multiple falco lists can have the same name.

        **Arguments**
            - None

        **Success Return Value**
            A JSON object representing the list of falco lists.
        '''
        res = self.http.get(self.url + '/api/secure/falco/lists/summaries', headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def get_falco_lists_group(self, name):
        '''**Description**
            Retrieve a group of all falco lists having the given name. This is used
            to show how a base list is modified by later lists that override/append
            to the list.

        **Arguments**
            - name: the name of the falco lists group

        **Success Return Value**
            A JSON object representing the list of falco lists.
        '''
        res = self.http.get(self.url + '/api/secure/falco/lists/groups?name={}'.format(name), headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def get_falco_list_id(self, id):
        '''**Description**
            Retrieve info about a single falco list

        **Arguments**
            - id: the id of the falco list

        **Success Return Value**
            A JSON object representing the falco list.
        '''
        res = self.http.get(self.url + '/api/secure/falco/lists/{}'.format(id), headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def add_falco_list(self, name, items, append=False):
        '''**Description**
            Create a new list

        **Arguments**
            - name: A name for this object. Should exactly be the value of the "list" property of the yaml object.
            - items: the array of items as represented in the yaml List.

        **Success Return Value**
            A JSON object representing the falco list.
        '''
        flist = {
            "name": name,
            "items": {
                "items": items
            },
            "append": append
        }
        res = self.http.post(self.url + '/api/secure/falco/lists', data=json.dumps(flist), headers=self.hdrs,
                             verify=self.ssl_verify)
        return self._request_result(res)

    def update_falco_list(self, id, items):
        '''**Description**
            Update info associated with a list

        **Arguments**
            - id: The rule id
            - items: the array of items as represented in the yaml List.

        **Success Return Value**
            A JSON object representing the list.
        '''
        ok, res = self.get_falco_list_id(id)
        if not ok:
            return [False, res]
        flist = res
        flist['items']['items'] = items

        res = self.http.put(self.url + '/api/secure/falco/lists/{}'.format(id), data=json.dumps(flist),
                            headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def delete_falco_list(self, id):
        '''**Description**
            Delete the list with given id.

        **Arguments**
            - id: The list id

        **Success Return Value**
            A JSON object representing the list.
        '''
        res = self.http.delete(self.url + '/api/secure/falco/lists/{}'.format(id), headers=self.hdrs,
                               verify=self.ssl_verify)
        return self._request_result(res)

    def add_compliance_task(self, name, module_name='docker-bench-security', schedule='06:00:00Z/PT12H', scope=None,
                            enabled=True):
        '''**Description**
            Add a new compliance task.

        **Arguments**
            - name: The name of the task e.g. 'Check Docker Compliance'.
            - module_name: The name of the module that implements this task. Separate from task name in case you want to use the same module to run separate tasks with different scopes or schedules. [ 'docker-bench-security', 'kube-bench' ]
            - schedule: The frequency at which this task should run. Expressed as an `ISO 8601 Duration <https://en.wikipedia.org/wiki/ISO_8601#Durations>`_
            - scope: The agent will only run the task on hosts matching this scope or on hosts where containers match this scope.
            - enabled: Whether this task should actually run as defined by its schedule.

        **Success Return Value**
            A JSON representation of the compliance task.
        '''
        task = {
            "id": None,
            "name": name,
            "moduleName": module_name,
            "enabled": enabled,
            "scope": scope,
            "schedule": schedule
        }
        res = self.http.post(self.url + '/api/complianceTasks', data=json.dumps(task), headers=self.hdrs,
                             verify=self.ssl_verify)
        return self._request_result(res)

    def list_compliance_tasks(self):
        '''**Description**
            Get the list of all compliance tasks.

        **Arguments**
            - None

        **Success Return Value**
            A JSON list with the representation of each compliance task.
        '''
        res = self.http.get(self.url + '/api/complianceTasks', headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def get_compliance_task(self, id):
        '''**Description**
            Get a compliance task.

        **Arguments**
            - id: the id of the compliance task to get.

        **Success Return Value**
            A JSON representation of the compliance task.
        '''
        res = self.http.get(self.url + '/api/complianceTasks/{}'.format(id), headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def update_compliance_task(self, id, name=None, module_name=None, schedule=None, scope=None, enabled=None):
        '''**Description**
            Update an existing compliance task.

        **Arguments**
            - id: the id of the compliance task to be updated.
            - name: The name of the task e.g. 'Check Docker Compliance'.
            - module_name: The name of the module that implements this task. Separate from task name in case you want to use the same module to run separate tasks with different scopes or schedules. [ 'docker-bench-security', 'kube-bench' ]
            - schedule: The frequency at which this task should run. Expressed as an `ISO 8601 Duration <https://en.wikipedia.org/wiki/ISO_8601#Durations>`_
            - scope: The agent will only run the task on hosts matching this scope or on hosts where containers match this scope.
            - enabled: Whether this task should actually run as defined by its schedule.

        **Success Return Value**
            A JSON representation of the compliance task.
        '''
        ok, res = self.get_compliance_task(id)
        if not ok:
            return ok, res

        task = res
        options = {
            'name': name,
            'moduleName': module_name,
            'schedule': schedule,
            'scope': scope,
            'enabled': enabled
        }
        task.update({k: v for k, v in options.items() if v is not None})
        res = self.http.put(self.url + '/api/complianceTasks/{}'.format(id), data=json.dumps(task), headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def delete_compliance_task(self, id):
        '''**Description**
            Delete the compliance task with the given id

        **Arguments**
            - id: the id of the compliance task to delete
        '''
        res = self.http.delete(self.url + '/api/complianceTasks/{}'.format(id), headers=self.hdrs,
                               verify=self.ssl_verify)
        if not self._checkResponse(res):
            return False, self.lasterr

        return True, None

    def list_compliance_results(self, limit=50, direction=None, cursor=None, filter=""):
        '''**Description**
            Get the list of all compliance tasks runs.

        **Arguments**
            - limit: Maximum number of alerts in the response.
            - direction: the direction (PREV or NEXT) that determines which results to return in relation to cursor.
            - cursor: An opaque string representing the current position in the list of alerts. It's provided in the 'responseMetadata' of the list_alerts response.
            - filter: an optional case insensitive filter used to match against the completed task name and return matching results.

        **Success Return Value**
            A JSON list with the representation of each compliance task run.
        '''
        url = "{url}/api/complianceResults?cursor{cursor}&filter={filter}&limit={limit}{direction}".format(
            url=self.url,
            limit=limit,
            direction="&direction=%s" % direction if direction else "",
            cursor="=%d" % cursor if cursor is not None else "",
            filter=filter)
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def get_compliance_results(self, id):
        '''**Description**
            Retrieve the details for a specific compliance task run result.

        **Arguments**
            - id: the id of the compliance task run to get.

        **Success Return Value**
            A JSON representation of the compliance task run result.
        '''
        res = self.http.get(self.url + '/api/complianceResults/{}'.format(id), headers=self.hdrs,
                            verify=self.ssl_verify)
        return self._request_result(res)

    def get_compliance_results_csv(self, id):
        '''**Description**
            Retrieve the details for a specific compliance task run result in csv.

        **Arguments**
            - id: the id of the compliance task run to get.

        **Success Return Value**
            A CSV representation of the compliance task run result.
        '''
        res = self.http.get(self.url + '/api/complianceResults/{}/csv'.format(id), headers=self.hdrs,
                            verify=self.ssl_verify)
        if not self._checkResponse(res):
            return False, self.lasterr

        return True, res.text

    def list_commands_audit(self, from_sec=None, to_sec=None, scope_filter=None, command_filter=None, limit=100,
                            offset=0, metrics=[]):
        '''**Description**
            List the commands audit.

        **DEPRECATED**: Use sdcclient.secure.ActivityAuditClientV1 instead. This is maintained for old on-prem versions, but will be removed over time.

        **Arguments**
            - from_sec: the start of the timerange for which to get commands audit.
            - end_sec: the end of the timerange for which to get commands audit.
            - scope_filter: this is a SysdigMonitor-like filter (e.g 'container.image=ubuntu'). When provided, commands are filtered by their scope, so only a subset will be returned (e.g. 'container.image=ubuntu' will provide only commands that have happened on an ubuntu container).
            - command_filter: this is a SysdigMonitor-like filter (e.g. command.comm="touch"). When provided, commands are filtered by some of their properties. Currently the supported set of filters is command.comm, command.cwd, command.pid, command.ppid, command.uid, command.loginshell.id, command.loginshell.distance
            - limit: Maximum number of commands in the response.
            - metrics: A list of metric values to include in the return.

        **Success Return Value**
            A JSON representation of the commands audit.
        '''
        if to_sec is None:
            to_sec = time.time()
        if from_sec is None:
            from_sec = to_sec - (24 * 60 * 60)  # 1 day

        url = "{url}/api/commands?from={frm}&to={to}&offset={offset}&limit={limit}{scope}{commandFilter}{metrics}".format(
            url=self.url,
            offset=offset,
            limit=limit,
            frm=int(from_sec * 10 ** 6),
            to=int(to_sec * 10 ** 6),
            scope="&scopeFilter=" + scope_filter if scope_filter else "",
            commandFilter="&commandFilter=" + command_filter if command_filter else "",
            metrics="&metrics=" + json.dumps(metrics) if metrics else "")
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def get_command_audit(self, id, metrics=[]):
        '''**Description**
            Get a command audit.

        **DEPRECATED**: Use sdcclient.secure.ActivityAuditClientV1 instead. This is maintained for old on-prem versions, but will be removed over time.

        **Arguments**
            - id: the id of the command audit to get.

        **Success Return Value**
            A JSON representation of the command audit.
        '''
        url = "{url}/api/commands/{id}?from=0&to={to}{metrics}".format(
            url=self.url,
            id=id,
            to=int(time.time() * 10 ** 6),
            metrics="&metrics=" + json.dumps(metrics) if metrics else "")
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def list_image_profiles(self):
        '''**Description**
            List the current set of image profiles.

        **Arguments**
            - None

        **Success Return Value**
            A JSON object containing the details of each profile.

        '''
        url = "{url}/api/v1/profiling/profileGroups/0/profiles".format(
            url=self.url
        )

        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        return self._request_result(res)

    def get_image_profile(self, profileId):
        '''**Description**
            Find the image profile with a (partial) profile ID <profileId> and return its json description.

        **Arguments**
            - name: the name of the image profile to fetch

        **Success Return Value**
            A JSON object containing the description of the image profile. If there is no image profile with
            the given name, returns False. Moreover, it could happen that more than one profile IDs have a collision.
            It is due to the fact that a partial profile ID can be passed and interpreted; in this case a set of
            collision profiles is returned, and the full complete ID string is printed. In this case, it returns
            false.

        '''

        # RETRIEVE ALL THE IMAGE PROFILES
        ok, image_profiles = self.list_image_profiles()

        if not ok:
            return [False, self.lasterr]

        '''
        The content of the json stored in the image_profiles dictionary:

        {
            "offset": 0,
            "limit": 99,
            "canLoadMore": false,
            "profiles": [
            ...
            ]
        }
        '''

        matched_profiles = self.__get_matched_profileIDs(profileId, image_profiles['profiles'])

        # Profile ID not found
        if len(matched_profiles) == 0:
            return [False, "No profile with ID {}".format(profileId)]

        # Principal workflow. Profile ID found
        elif len(matched_profiles) == 1:
            # Matched id. Return information
            url = "{url}/api/v1/profiling/profiles/{profileId}".format(
                url=self.url,
                profileId=matched_profiles[0]['profileId']
            )

            res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
            return self._request_result(res)

        # Collision detected. The full profile IDs are returned
        elif len(matched_profiles) >= 2:
            return [False, matched_profiles]

    def __get_matched_profileIDs(self, requested_profile, profile_list):
        '''
        **Description**
            Helper function for  retrieving the list of matching profile

        **Arguments**
            - the requested profile Id (string)
            - List of dictionary, where each dictionary contains the profile information

        **Success Return Value**
            List of dictionary, where each dictionary represents a profile with the ID prefix substring
            matching the requested one

        **Content structure of the profile_list parameter**
        This array of profiles contains all the relevant information. For the purposes of this function, only
        the profileId field is relevant.

        [
            {
            "profileGroupId": 0,
            "profileId": "00000000000000000000000000000000000000000000",
            "profileVersion": 0,
            "profileName": "AAA/BBB:XYZ@0000000000000000000000",
            "imageId": "00000000000000000000000000000000000000000000",
            "imageName": "AAA/BBB:XYZ",
            "processesProposal": {
                "subcategories": [
                                    {
                                    "name": "process",
                                    "ruleName": "process - 00000000000000000000000000000000000000000000",
                                    "ruleType": "PROCESS",
                                    "score": 000
                                    }
                                ],
                                "score": 000
            },
            "fileSystemProposal": {
                "subcategories": [
                                    {
                                    "name": "filesystem",
                                    "ruleName": "filesystem - 00000000000000000000000000000000000000000000",
                                    "ruleType": "FILESYSTEM",
                                    "score": 000
                                    }
                                ],
                                "score": 000
            },
            "syscallProposal": {
                "subcategories": [
                                    {
                                    "name": "syscalls",
                                    "ruleName": "syscalls - 00000000000000000000000000000000000000000000",
                                    "ruleType": "SYSCALL",
                                    "score": 000
                                    }
                                ],
                                "score": 000
            },
            "networkProposal": {
                "subcategories": [
                                    {
                                    "name": "network",
                                    "ruleName": "network - 00000000000000000000000000000000000000000000",
                                    "ruleType": "NETWORK",
                                    "score": 000
                                    }
                                ],
                                "score": 000
            },
            "containerImagesProposal": {
                "subcategories": [
                                    {
                                    "name": "container image",
                                    "ruleName": "container image - 00000000000000000000000000000000000000000000",
                                    "ruleType": "CONTAINER",
                                    "score": 0
                                    }
                                ],
                                "score": 0
            },
            "status": "STATUS_VALUE",
            "score": 000
            },
            ...
        ]
        '''

        matched_profiles = []

        request_len = len(requested_profile)
        for profile in profile_list:

            # get the length of the substring to match
            str_len_match = min(len(profile), request_len)

            if profile['profileId'][0:str_len_match] == requested_profile[0:str_len_match]:
                matched_profiles.append(profile)

        return matched_profiles



import datetime

from sdcclient._common import _SdcCommon


class ActivityAuditDataSource:
    CMD = "command"
    NET = "connection"
    KUBE_EXEC = "kubernetes"
    FILE = "fileaccess"


_seconds_to_nanoseconds = 10 ** 9


class ActivityAuditClientV1(_SdcCommon):
    def __init__(self, token="", sdc_url='https://secure.sysdig.com', ssl_verify=True, custom_headers=None):
        super(ActivityAuditClientV1, self).__init__(token, sdc_url, ssl_verify, custom_headers)
        self.product = "SDS"

    def list_events(self, from_date=None, to_date=None, scope_filter=None, limit=0,
                    data_sources=None):
        """
        List the events in the Activity Audit.

        Args:
            from_date (datetime.datetime): the start of the time range from which to get events. The default value is yesterday.
            to_date (datetime.datetime): the end of the time range from which to get events. The default value is now.
            scope_filter (List): a list of Sysdig Monitor-like filter (e.g `processName in ("ubuntu")`).
            limit (int): max number of events to retrieve. A limit of 0 or negative will retrieve all events.
            data_sources (List): a list of data sources to retrieve events from. None or an empty list retrieves all events.

        Examples:
            >>> client = ActivityAuditClientV1(token=SECURE_TOKEN)
            >>>
            >>> now = datetime.datetime.now()
            >>> three_days_ago = now - datetime.timedelta(days=3)
            >>> max_event_number_retrieved = 50
            >>> data_sources = [ActivityAuditDataSource.CMD, ActivityAuditDataSource.KUBE_EXEC]
            >>>
            >>> ok, events = client.list_events(from_date=three_days_ago,
            >>>                                 to_date=now,
            >>>                                 limit=max_event_number_retrieved,
            >>>                                 data_sources=data_sources)

        Returns:
            A list of event objects from the Activity Audit.
        """
        number_of_events_per_query = 50

        if from_date is None:
            from_date = datetime.datetime.now() - datetime.timedelta(days=1)
        if to_date is None:
            to_date = datetime.datetime.now()

        filters = scope_filter if scope_filter else []
        if data_sources:
            quoted_data_sources = [f'"{data_source}"' for data_source in data_sources]
            data_source_filter = f'type in ({",".join(quoted_data_sources)})'
            filters.append(data_source_filter)

        query_params = {
            "from": int(from_date.timestamp()) * _seconds_to_nanoseconds,
            "to": int(to_date.timestamp()) * _seconds_to_nanoseconds,
            "limit": number_of_events_per_query,
            "filter": " and ".join(filters),
        }

        res = self.http.get(self.url + '/api/v1/activityAudit/events', headers=self.hdrs, verify=self.ssl_verify,
                            params=query_params)
        ok, res = self._request_result(res)
        if not ok:
            return False, res

        events = []

        # Pagination required by Secure API
        while "page" in res and \
                "total" in res["page"] and \
                res["page"]["total"] > number_of_events_per_query:
            events = events + res["data"]

            if 0 < limit < len(events):
                events = events[0:limit - 1]
                break

            paginated_query_params = {
                "limit": number_of_events_per_query,
                "filter": " and ".join(filters),
                "cursor": res["page"]["prev"]
            }

            res = self.http.get(self.url + '/api/v1/activityAudit/events', headers=self.hdrs, verify=self.ssl_verify,
                                params=paginated_query_params)
            ok, res = self._request_result(res)
            if not ok:
                return False, res
        else:
            events = events + res["data"]

        return True, events

    def list_trace(self, traceable_event):
        """
        Lists the events from an original traceable event.

        Args:
            traceable_event(object): an event retrieved from the list_events method. The event must be traceable,
                                     this is, it must have the "traceable" key as true.

        Examples:
            >>> client = ActivityAuditClientV1(token=SECURE_TOKEN)
            >>>
            >>> ok, events = client.list_events()
            >>> if not ok:
            >>>     return
            >>> traceable_events = [event for event in events if event["traceable"]]
            >>>
            >>> ok, trace = client.list_trace(traceable_events[0])
            >>> if not ok:
            >>>     return
            >>>
            >>> for event in trace:
            >>>     print(event)

        Returns:
            All the related events that are the trace of the given event.
        """
        if not traceable_event or not traceable_event["traceable"]:
            return False, "a traceable event must be provided"

        endpoint = f'/api/v1/activityAudit/events/{traceable_event["type"]}/{traceable_event["id"]}/trace'
        res = self.http.get(self.url + endpoint, headers=self.hdrs, verify=self.ssl_verify)
        ok, res = self._request_result(res)
        if not ok:
            return False, res
        return True, res["data"]



import base64
import json
import re
import time
from datetime import datetime
from warnings import warn

from requests.exceptions import RetryError
from requests_toolbelt.multipart.encoder import MultipartEncoder

from sdcclient.secure.scanning import ScanningAlertsClientV1

try:
    from urllib.parse import quote_plus, unquote_plus
except ImportError:
    from urllib import quote_plus, unquote_plus

from sdcclient._common import _SdcCommon


class SdScanningClient(ScanningAlertsClientV1, _SdcCommon):
    def __init__(self, token="", sdc_url='https://secure.sysdig.com', ssl_verify=True, custom_headers=None):
        super(SdScanningClient, self).__init__(token, sdc_url, ssl_verify, custom_headers)
        self.product = "SDS"

    def add_image(self, image, force=False, dockerfile=None, annotations={}, autosubscribe=True):
        '''**Description**
            Add an image to the scanner

        **Arguments**
            - image: Input image can be in the following formats: registry/repo:tag
            - dockerfile: The contents of the dockerfile as a str.
            - annotations: A dictionary of annotations {str: str}.
            - autosubscribe: Should active the subscription to this image?

        **Success Return Value**
            A JSON object representing the image that was added.
        '''
        itype = self._discover_inputimage_format(image)
        if itype != 'tag':
            return [False, "can only add a tag"]

        payload = {}
        if dockerfile:
            payload['dockerfile'] = base64.b64encode(dockerfile.encode()).decode("utf-8")
        payload['tag'] = image
        if annotations:
            payload['annotations'] = annotations

        url = "{base_url}/api/scanning/v1/anchore/images?autosubscribe={autosubscribe}{force}".format(
                base_url=self.url,
                autosubscribe=str(autosubscribe),
                force="&force=true" if force else "")

        res = self.http.post(url, data=json.dumps(payload), headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def get_image(self, image, show_history=False):
        '''**Description**
            Find the image with the tag <image> and return its json description

        **Arguments**
            - image: Input image can be in the following formats: registry/repo:tag

        **Success Return Value**
            A JSON object representing the image.
        '''
        itype = self._discover_inputimage_format(image)
        if itype not in ['tag', 'imageid', 'imageDigest']:
            return [False, "cannot use input image string: no discovered imageDigest"]

        params = {}
        params['history'] = str(show_history and itype not in ['imageid', 'imageDigest']).lower()
        if itype == 'tag':
            params['fulltag'] = image

        url = self.url + "/api/scanning/v1/anchore/images"
        url += {
                'imageid':     '/by_id/{}'.format(image),
                'imageDigest': '/{}'.format(image)
        }.get(itype, '')

        res = self.http.get(url, params=params, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def list_images(self):
        '''**Description**
            List the current set of images in the scanner.

        **Arguments**
            - None

        **Success Return Value**
            A JSON object containing all the images.
        '''
        url = self.url + "/api/scanning/v1/anchore/images"
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def list_image_tags(self):
        """
        Lists the current set of image tags in the scanner.

        Returns: A JSON object containing all the image tags.
        """
        url = self.url + "/api/scanning/v1/anchore/summaries/imagetags"
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def list_whitelisted_cves(self):
        '''**Description**
            List the whitelisted global CVEs.

        **Arguments**
            - None

        **Success Return Value**
            A JSON object containing all the whitelisted CVEs.

        **Deprecated**
            This method has been deprecated since the API has changed. Use the
            list_vulnerability_exception_bundles and get_vulnerability_exception_bundle methods.
        '''
        warn("list_whitelisted_cves has been deprecated and doesn't work properly, please use the "
             "list_vulnerability_exception_bundles and get_vulnerability_exception_bundle methods",
             DeprecationWarning, 3)
        url = self.url + "/api/scanning/v1/whitelists/global?bundle=default"
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def query_image_content(self, image, content_type=""):
        '''**Description**
            Find the image with the tag <image> and return its content.

        **Arguments**
            - image: Input image can be in the following formats: registry/repo:tag
            - content_type: The content type can be one of the following types:
                - os: Operating System Packages
                - files: Files
                - npm: Node.JS NPM Module
                - gem: Ruby GEM
                - python: Python modules
                - java: Java packages

        **Success Return Value**
            A JSON object representing the image content.
        '''
        content_type = content_type.lower()
        supported_types = ["os", "files", "npm", "gem", "python", "java"]
        if content_type not in supported_types:
            return False, f"unsupported type provided: {content_type}, must be one of {supported_types}"

        return self._query_image(image, query_group='content', query_type=content_type)

    def query_image_metadata(self, image, metadata_type=""):
        '''**Description**
            Find the image with the tag <image> and return its metadata.

        **Arguments**
            - image: Input image can be in the following formats: registry/repo:tag
            - metadata_type: The metadata type can be one of the types returned by running without a type specified

        **Success Return Value**
            A JSON object representing the image metadata.
        '''
        return self._query_image(image, query_group='metadata', query_type=metadata_type)

    def query_image_vuln(self, image, vuln_type="", vendor_only=True):
        '''**Description**
            Find the image with the tag <image> and return its vulnerabilities.

        **Arguments**
            - image: Input image can be in the following formats: registry/repo:tag
            - vuln_type: Vulnerability type can be one of the following types:
                - os: CVE/distro vulnerabilities against operating system packages

        **Success Return Value**
            A JSON object representing the image vulnerabilities.
        '''
        return self._query_image(image, query_group='vuln', query_type=vuln_type, vendor_only=vendor_only)

    def query_images_by_vulnerability(self, vulnerability_id, namespace=None, package=None, severity=None,
                                      vendor_only=True):
        '''**Description**
            Search system for images with the given vulnerability ID present

        **Arguments**
            - vulnerability_id: Search for images vulnerable to this vulnerability ID (e.g. CVE-1999-0001)
            - namespace: Filter results to images with vulnerable packages in the given namespace (e.g. debian:9)
            - package: Filter results to images with the given vulnerable package name (e.g. sed)
            - severity: Filter results to images with the given vulnerability severity (e.g. Medium)
            - vendor_only: Only show images with vulnerabilities explicitly deemed applicable by upstream OS vendor, if present

        **Success Return Value**
            A JSON object representing the images.
        '''
        url = "{base_url}/api/scanning/v1/anchore/query/images/by_vulnerability?vulnerability_id={vulnerability_id}{namespace}{package}{severity}&vendor_only={vendor_only}".format(
                base_url=self.url,
                vulnerability_id=vulnerability_id,
                namespace="&namespace={}".format(namespace) if namespace else "",
                package="&affected_package={}".format(package) if package else "",
                severity="&severity={}".format(severity) if severity else "",
                vendor_only=vendor_only)

        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def query_images_by_package(self, name, version=None, package_type=None):
        '''**Description**
            Search system for images with the given package installed

        **Arguments**
            - name: Search for images with this package name (e.g. sed)
            - version: Filter results to only packages with given version (e.g. 4.4-1)
            - package-type: Filter results to only packages of given type (e.g. dpkg)

        **Success Return Value**
            A JSON object representing the images.
        '''
        url = "{base_url}/api/scanning/v1/anchore/query/images/by_package?name={name}{version}{package_type}".format(
                base_url=self.url,
                name=name,
                version="&version={}".format(version) if version else "",
                package_type="&package_type={}".format(package_type) if package_type else "")

        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def _query_image(self, image, query_group="", query_type="", vendor_only=True):
        if not query_group:
            raise Exception("need to specify a query group")

        _, _, image_digest = self._discover_inputimage(image)
        if not image_digest:
            return [False, "cannot use input image string (no discovered imageDigest)"]

        url = "{base_url}/api/scanning/v1/anchore/images/{image_digest}/{query_group}/{query_type}{vendor_only}".format(
                base_url=self.url,
                image_digest=image_digest,
                query_group=query_group,
                query_type=query_type if query_type else '',
                vendor_only="?vendor_only={}".format(vendor_only) if query_group == 'vuln' else '')

        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def delete_image(self, image, force=False):
        '''**Description**
            Delete image from the scanner.

        **Arguments**
            - None
        '''
        _, _, image_digest = self._discover_inputimage(image)
        if not image_digest:
            return [False, "cannot use input image string: no discovered imageDigest"]

        url = self.url + "/api/scanning/v1/anchore/images/" + image_digest
        res = self.http.delete(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def check_image_evaluation(self, image, show_history=False, detail=False, tag=None, policy=None):
        '''**Description**
            Check the latest policy evaluation for an image

        **Arguments**
            - image: Input image can be in the following formats: registry/repo:tag
            - show_history: Show all previous policy evaluations
            - detail: Show detailed policy evaluation report
            - tag: Specify which TAG is evaluated for a given image ID or Image Digest
            - policy: Specify which POLICY to use for evaluate (defaults currently active policy)

        **Success Return Value**
            A JSON object representing the evaluation status.
        '''
        itype, _, image_digest = self._discover_inputimage(image)
        if not image_digest:
            return [False, "could not get image record from anchore"]
        if not tag and itype != 'tag':
            return [False, "input image name is not a tag, and no --tag is specified"]

        thetag = tag if tag else image

        url = "{base_url}/api/scanning/v1/anchore/images/{image_digest}/check?history={history}&detail={detail}&tag={tag}{policy_id}"
        url = url.format(
                base_url=self.url,
                image_digest=image_digest,
                history=str(show_history).lower(),
                detail=str(detail).lower(),
                tag=thetag,
                policy_id=("&policyId=%s" % policy) if policy else "")

        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def get_pdf_report(self, image, tag=None, date=None):
        '''**Description**
            Get a pdf report of one image

        **Arguments**
            - image: Input image can be in the following formats: registry/repo:tag
            - tag: Specify which TAG is evaluated for a given image ID or Image Digest
            - date: date for the report of interest, the format is 'YYYY-MM-DDTHH:MM:SSZ',
                    if not provided the latest report will be returned

        **Success Return Value**
            The pdf content
        '''
        image_type, _, image_digest = self._discover_inputimage(image)
        if not image_digest:
            return [False, "could not get image record from anchore"]
        if not tag and image_type != 'tag':
            return [False, "input image name is not a tag"]
        image_tag = tag if tag else image

        url = "{base_url}/api/scanning/v1/images/{image_digest}/report?tag={tag}{at}".format(
                base_url=self.url,
                image_digest=image_digest,
                tag=image_tag,
                at=("&at=%s" % date) if date else "")

        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.content]

    def get_latest_pdf_report_by_digest(self, image_digest, full_tag=None):
        '''**Description**
            Get the latest pdf report of one image digest

        **Arguments**
            - image_digest: Input image digest should be in the following formats: sha256:134dhgfd65765
            - tag: Specify which FULLTAG is evaluated for a given Image Digest: docker.io/alpine:latest

        **Success Return Value**
            The pdf content
        '''
        url = "{base_url}/api/scanning/v1/images/{image_digest}/report?tag={tag}".format(
                base_url=self.url,
                image_digest=image_digest,
                tag=full_tag)

        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.content]

    def import_image(self, infile, image_id, digest_id, image_name, sync=False):
        '''**Description**
            Import an image archive

        **Arguments**
           - infile: An image archive file
           - sync: Whether the import should be synchronous or asynchronous

        **Success Return Value**
            If synchronous, A JSON object representing the image that was imported.

        '''
        try:
            m = MultipartEncoder(
                    fields={'archive_file': (infile, open(infile, 'rb'), 'text/plain')}
            )
            if sync:
                url = self.url + "/api/scanning/v1/anchore/import/images"
            else:
                url = self.url + "/api/scanning/v1/import/images"

            headers = {'Authorization': 'Bearer ' + self.token, 'Content-Type': m.content_type,
                       'imageId':       image_id, 'digestId': digest_id, 'imageName': image_name}
            res = self.http.post(url, data=m, headers=headers, verify=self.ssl_verify)
            if not self._checkResponse(res):
                return [False, self.lasterr]

            return [True, res.json() if sync else res.content]

        except Exception as err:
            print(err)

    def get_anchore_users_account(self):
        '''**Description**
            Get the anchore user account.

        **Arguments**
            - None

        **Success Return Value**
            A JSON object containing user account information.
        '''
        url = self.url + "/api/scanning/v1/account"
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def get_image_scan_result_by_id(self, image_id, full_tag_name, detail):
        '''**Description**
            Get the anchore image scan result for an image id.

        **Arguments**
            - image_id: Docker image id of the image whose scan result is to be fetched.
            - full_tag_name: The complete tag name of the image for e.g. docker.io/alpine:3.10.
            - detail: Boolean to indicate whether full scan result API is needed.

        **Success Return Value**
            A JSON object containing pass/fail status of image scan policy.
        '''
        url = "{base_url}/api/scanning/v1/anchore/images/by_id/{image_id}/check?tag={full_tag_name}&detail={detail}".format(
                base_url=self.url,
                image_id=image_id,
                full_tag_name=full_tag_name,
                detail=detail)
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def add_registry(self, registry, registry_user, registry_pass, insecure=False, registry_type="docker_v2",
                     validate=True):
        '''**Description**
            Add image registry

        **Arguments**
            - registry: Full hostname/port of registry. Eg. myrepo.example.com:5000
            - registry_user: Username
            - registry_pass: Password
            - insecure: Allow connection to registry without SSL cert checks (ex: if registry uses a self-signed SSL certificate)
            - registry_type: Specify the registry type. 'docker_v2' and 'awsecr' are supported (default='docker_v2')
            - validate: If set to 'False' will not attempt to validate registry/creds on registry add

        **Success Return Value**
            A JSON object representing the registry.
        '''
        registry_types = ['docker_v2', 'awsecr']
        if registry_type and registry_type not in registry_types:
            return [False, "input registry type not supported (supported registry_types: " + str(registry_types)]
        if self._registry_string_is_valid(registry):
            return [False,
                    "input registry name cannot contain '/' characters - valid registry names are of the form <host>:<port> where :<port> is optional"]

        if not registry_type:
            registry_type = self._get_registry_type(registry)

        payload = {
                'registry':        registry,
                'registry_user':   registry_user,
                'registry_pass':   registry_pass,
                'registry_type':   registry_type,
                'registry_verify': not insecure}
        url = "{base_url}/api/scanning/v1/anchore/registries?validate={validate}".format(
                base_url=self.url,
                validate=validate)

        res = self.http.post(url, data=json.dumps(payload), headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def update_registry(self, registry, registry_user, registry_pass, insecure=False, registry_type="docker_v2",
                        validate=True):
        '''**Description**
            Update an existing image registry.

        **Arguments**
            - registry: Full hostname/port of registry. Eg. myrepo.example.com:5000
            - registry_user: Username
            - registry_pass: Password
            - insecure: Allow connection to registry without SSL cert checks (ex: if registry uses a self-signed SSL certificate)
            - registry_type: Specify the registry type. 'docker_v2' and 'awsecr' are supported (default='docker_v2')
            - validate: If set to 'False' will not attempt to validate registry/creds on registry add

        **Success Return Value**
            A JSON object representing the registry.
        '''
        if self._registry_string_is_valid(registry):
            return [False,
                    "input registry name cannot contain '/' characters - valid registry names are of the form <host>:<port> where :<port> is optional"]

        payload = {
                'registry':        registry,
                'registry_user':   registry_user,
                'registry_pass':   registry_pass,
                'registry_type':   registry_type,
                'registry_verify': not insecure}
        url = "{base_url}/api/scanning/v1/anchore/registries/{registry}?validate={validate}".format(
                base_url=self.url,
                registry=registry,
                validate=validate)

        res = self.http.put(url, data=json.dumps(payload), headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def delete_registry(self, registry):
        '''**Description**
            Delete an existing image registry

        **Arguments**
            - registry: Full hostname/port of registry. Eg. myrepo.example.com:5000
        '''
        # do some input string checking
        if re.match(".*\\/.*", registry):
            return [False,
                    "input registry name cannot contain '/' characters - valid registry names are of the form <host>:<port> where :<port> is optional"]

        url = self.url + "/api/scanning/v1/anchore/registries/" + registry
        res = self.http.delete(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def list_registry(self):
        '''**Description**
            List all current image registries

        **Arguments**
            - None

        **Success Return Value**
            A JSON object representing the list of registries.
        '''
        url = self.url + "/api/scanning/v1/anchore/registries"
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def get_registry(self, registry):
        '''**Description**
            Find the registry and return its json description

        **Arguments**
            - registry: Full hostname/port of registry. Eg. myrepo.example.com:5000

        **Success Return Value**
            A JSON object representing the registry.
        '''
        if self._registry_string_is_valid(registry):
            return [False,
                    "input registry name cannot contain '/' characters - valid registry names are of the form <host>:<port> where :<port> is optional"]

        url = self.url + "/api/scanning/v1/anchore/registries/" + registry
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def _get_registry_type(self, registry):
        if re.match("[0-9]+\\.dkr\\.ecr\\..*\\.amazonaws\\.com", registry):
            return "awsecr"
        return "docker_v2"

    def _registry_string_is_valid(self, registry):
        return re.match(".*\\/.*", registry)

    def add_repo(self, repo, autosubscribe=True, lookuptag=None):
        '''**Description**
            Add a repository

        **Arguments**
            - repo: Input repository can be in the following formats: registry/repo
            - autosubscribe: If unset, instruct the engine to disable subscriptions for any discovered tags.
            - lookuptag: Specify a tag to use for repo tag scan if 'latest' tag does not exist in the repo.

        **Success Return Value**
            A JSON object representing the repo.
        '''
        url = "{base_url}/api/scanning/v1/anchore/repositories?repository={repo}&autosubscribe={autosubscribe}{lookuptag}".format(
                base_url=self.url,
                repo=repo,
                autosubscribe=autosubscribe,
                lookuptag="&lookuptag={}".format(lookuptag) if lookuptag else "")

        res = self.http.post(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def watch_repo(self, repo):
        '''**Description**
            Instruct engine to start automatically watching the repo for image updates

        **Arguments**
            - repo: Input repository can be in the following formats: registry/repo
        '''
        return self.activate_subscription('repo_update', repo)

    def unwatch_repo(self, repo):
        '''**Description**
            Instruct engine to stop automatically watching the repo for image updates

        **Arguments**
            - repo: Input repository can be in the following formats: registry/repo
        '''
        return self.deactivate_subscription('repo_update', repo)

    def delete_repo(self, repo):
        '''**Description**
            Delete a repository from the watch list (does not delete already analyzed images)

        **Arguments**
            - repo: Input repository can be in the following formats: registry/repo
        '''
        return self.delete_subscription('repo_update', repo)

    def list_repos(self):
        '''**Description**
            List added repositories

        **Arguments**
            - None

        **Success Return Value**
            A JSON object representing the list of repositories.
        '''
        return self.get_subscriptions("repo_update")

    def get_repo(self, repo):
        '''**Description**
            Get a repository

        **Arguments**
            - repo: Input repository can be in the following formats: registry/repo

        **Success Return Value**
            A JSON object representing the registry.
        '''
        return self.get_subscriptions("repo_update", repo)

    def add_policy(self, name, rules, comment="", bundleid=None):
        '''**Description**
            Create a new policy

        **Arguments**
            - name: The name of the policy.
            - rules: A list of Anchore PolicyRule elements (while creating/updating a policy, new rule IDs will be created backend side)
            - comment: A human-readable description.
            - bundleid: Target bundle. If not specified, the currently active bundle will be used.

        **Success Return Value**
            A JSON object containing the policy description.
        '''
        policy = {
                'name':    name,
                'comment': comment,
                'rules':   rules,
                'version': '1_0'
        }
        if bundleid:
            policy['policyBundleId'] = bundleid

        url = self.url + '/api/scanning/v1/policies'
        data = json.dumps(policy)
        res = self.http.post(url, headers=self.hdrs, data=data, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def list_policy_bundles(self, detail=False):
        url = "{base_url}/api/scanning/v1/anchore/policies?detail={detail}".format(
                base_url=self.url,
                detail=str(detail))
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def list_policies(self, bundleid=None):
        '''**Description**
            List the current set of scanning policies.

        **Arguments**
            - bundleid: Target bundle. If not specified, the currently active bundle will be used.

        **Success Return Value**
            A JSON object containing the list of policies.
        '''
        url = self.url + '/api/scanning/v1/policies'
        if bundleid:
            url += '?bundleId=' + bundleid

        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def get_policy(self, policyid, bundleid=None):
        '''**Description**
            Retrieve the policy with the given id in the targeted policy bundle

        **Arguments**
            - policyid: Unique identifier associated with this policy.
            - bundleid: Target bundle. If not specified, the currently active bundle will be used.

        **Success Return Value**
            A JSON object containing the policy description.
        '''
        ok, policies = self.list_policies(bundleid)
        if not ok:
            return [ok, policies]

        for policy in policies:
            if policy["id"] == policyid:
                return [True, policy]

        return [False, "Policy not found"]

    def update_policy(self, policyid, policy_description):
        '''**Description**
            Update the policy with the given id

        **Arguments**
            - policyid: Unique identifier associated with this policy.
            - policy_description: A dictionary with the policy description.

        **Success Return Value**
            A JSON object containing the policy description.
        '''
        url = self.url + '/api/scanning/v1/policies/' + policyid
        data = json.dumps(policy_description)
        res = self.http.put(url, headers=self.hdrs, data=data, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def delete_policy(self, policyid, bundleid=None):
        '''**Description**
            Delete the policy with the given id in the targeted policy Bundle

        **Arguments**
            - policyid: Unique identifier associated with this policy.
            - policy_description: A dictionary with the policy description.
        '''
        url = self.url + '/api/scanning/v1/policies/' + policyid
        if bundleid:
            url += '?bundleId=' + bundleid

        res = self.http.delete(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.text]

    def add_alert(self, name, description=None, scope="", triggers={'failed': True, 'unscanned': True},
                  enabled=False, notification_channels=[]):
        '''
        Create a new alert
        **Warning**: `add_alert` is deprecated and will be removed soon, use
        `add_runtime_alert` or `add_repository_alert` from `ScanningAlertsClientV1` instead.

        Args:
            name(str): The name of the alert.
            description(str): The descprition of the alert.
            scope(str): An AND-composed string of predicates that selects the scope in which the alert will be applied. (like: 'host.domain = "example.com" and container.image != "alpine:latest"')
            tiggers(dict): A dict {str: bool} indicating wich triggers should be enabled/disabled. (default: {'failed': True, 'unscanned': True})
            enabled(bool): Whether this alert should actually be applied.
            notification_channels(list): A list of notification channel ids.

        Returns:
            A JSON object containing the alert description.

        .. deprecated::
            `add_alert` is deprecated and will be removed soon, use
            `add_runtime_alert` or `add_repository_alert` from `ScanningAlertsClientV1` instead.
        '''

        alert = {
                'name':                   name,
                'description':            description,
                'triggers':               triggers,
                'scope':                  scope,
                'enabled':                enabled,
                'autoscan':               True,
                'notificationChannelIds': notification_channels,
        }

        url = self.url + '/api/scanning/v1/alerts'
        data = json.dumps(alert)
        res = self.http.post(url, headers=self.hdrs, data=data, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def update_alert(self, alertid, alert_description):
        '''
        Update the alert with the given id.
        **Warning**: `update_alert` is deprecated and will be removed soon, use
        `update_runtime_alert` or `update_repository_alert` from `ScanningAlertsClientV1` instead.

        Args:
            alertid(str): Unique identifier associated with this alert.
            alert_description(str): A dictionary with the alert description.

        Returns:
            A JSON object containing the alert description.

        .. deprecated::
            `update_alert` is deprecated and will be removed soon, use
            `update_runtime_alert` or `update_repository_alert` from `ScanningAlertsClientV1` instead.
        '''
        url = self.url + '/api/scanning/v1/alerts/' + alertid
        data = json.dumps(alert_description)
        res = self.http.put(url, headers=self.hdrs, data=data, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def get_subscriptions(self, subscription_type=None, subscription_key=None):
        '''**Description**
            Get the list of subscriptions

        **Arguments**
            - subscription_type: Type of subscription. Valid options:
                - 'tag_update': Receive notification when new image is pushed
                - 'policy_eval': Receive notification when image policy status changes
                - 'vuln_update': Receive notification when vulnerabilities are added, removed or modified
                - 'repo_update': Receive notification when a repo is updated
            - subscription_key: Fully qualified name of tag to subscribe to. Eg. docker.io/library/alpine:latest
        '''
        url = self.url + "/api/scanning/v1/anchore/subscriptions/"
        if subscription_key or subscription_type:
            url += "?"
            if subscription_key:
                url += "subscription_key={}&".format(subscription_key)
            if subscription_type:
                url += "subscription_type={}".format(subscription_type)
        res = self.http.get(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def activate_subscription(self, subscription_type, subscription_key):
        '''**Description**
            Activate a subscription

        **Arguments**
            - subscription_type: Type of subscription. Valid options:
                - 'tag_update': Receive notification when new image is pushed
                - 'policy_eval': Receive notification when image policy status changes
                - 'vuln_update': Receive notification when vulnerabilities are added, removed or modified
                - 'repo_update': Receive notification when a repo is updated
            - subscription_key: Fully qualified name of tag to subscribe to. Eg. docker.io/library/alpine:latest
        '''
        return self._update_subscription(subscription_type, subscription_key, True)

    def deactivate_subscription(self, subscription_type, subscription_key):
        '''**Description**
            Deactivate a subscription

        **Arguments**
            - subscription_type: Type of subscription. Valid options:
                - 'tag_update': Receive notification when new image is pushed
                - 'policy_eval': Receive notification when image policy status changes
                - 'vuln_update': Receive notification when vulnerabilities are added, removed or modified
                - 'repo_update': Receive notification when a repo is updated
            - subscription_key: Fully qualified name of tag to subscribe to. Eg. docker.io/library/alpine:latest
        '''
        return self._update_subscription(subscription_type, subscription_key, False)

    def delete_subscription(self, subscription_type, subscription_key):
        '''**Description**
            Delete a subscription

        **Arguments**
            - subscription_type: Type of subscription. Valid options:
                - 'tag_update': Receive notification when new image is pushed
                - 'policy_eval': Receive notification when image policy status changes
                - 'vuln_update': Receive notification when vulnerabilities are added, removed or modified
                - 'repo_update': Receive notification when a repo is updated
            - subscription_key: Fully qualified name of tag to subscribe to. Eg. docker.io/library/alpine:latest
        '''
        try:
            url = self._subscription_url(subscription_type, subscription_key)
        except Exception as err:
            return [False, err]

        res = self.http.delete(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def _update_subscription(self, subscription_type, subscription_key, activate):
        try:
            url = self._subscription_url(subscription_type, subscription_key)
        except Exception as err:
            return [False, err]

        payload = {'active': activate, 'subscription_key': subscription_key, 'subscription_type': subscription_type}
        res = self.http.put(url, data=json.dumps(payload), headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def _subscription_url(self, subscription_type, subscription_key):
        ok, res = self.get_subscriptions(subscription_type, subscription_key)
        if not ok:
            raise Exception(res)

        if len(res) != 1:
            raise Exception("Subscription {} doesn't exist".format(subscription_key))
        id = res[0].get("subscription_id", None)
        if not id:
            raise Exception("Subscription malformed")

        return self.url + "/api/scanning/v1/anchore/subscriptions/" + id

    def list_subscription(self):
        '''**Description**
            List all subscriptions

        **Arguments**
            - None

        **Success Return Value**
            A JSON object representing the list of subscriptions.
        '''
        return self.get_subscriptions()

    def list_runtime(self, scope="", skip_policy_evaluation=True, start_time=None, end_time=None):
        '''**Description**
            List runtime containers

        **Arguments**
            - scope: An AND-composed string of predicates that selects the scope in which the alert will be applied. (like: 'host.domain = "example.com" and container.image != "alpine:latest"')
            - skip_policy_evaluation: If true, no policy evaluations will be triggered for the images.
            - start_time: Start of the time range (integer of unix time).
            - end_time: End of the time range (integer of unix time).

        **Success Return Value**
            A JSON object representing the list of runtime containers.
        '''
        containers = {
                'scope':                scope,
                'skipPolicyEvaluation': skip_policy_evaluation
        }
        if start_time or end_time:
            containers['time'] = {}
            containers['time']['from'] = int(start_time * 100000) if start_time else 0
            end_time = end_time if end_time else time.time()
            containers['time']['to'] = int(end_time * 1000000)

        url = self.url + '/api/scanning/v1/query/containers'
        data = json.dumps(containers)
        res = self.http.post(url, headers=self.hdrs, data=data, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def _discover_inputimage_format(self, input_string):
        itype = None

        if re.match("^sha256:[0-9a-fA-F]{64}", input_string):
            itype = 'imageDigest'
        elif re.match("[0-9a-fA-F]{64}", input_string):
            itype = 'imageid'
        else:
            itype = 'tag'

        return itype

    def _discover_inputimage(self, input_string):
        patt = re.match(".*(sha256:.*)", input_string)
        if patt:
            urldigest = quote_plus(patt.group(1))
            return "digest", input_string, urldigest

        try:
            digest = unquote_plus(str(input_string))
            for tpe in ["sha256", "local"]:
                patt = re.match(".*({}:.*)".format(tpe), digest)
                if patt:
                    return "imageDigest", input_string, input_string
        except Exception:
            pass

        urldigest = None
        ret_type = "tag"
        ok, ret = self.get_image(input_string)
        if ok:
            image_record = ret[0]
            urldigest = image_record.get('imageDigest', None)
            for image_detail in image_record.get('image_detail', []):
                if input_string == image_detail.get('imageId', ''):
                    ret_type = "imageid"
                    break

        return ret_type, input_string, urldigest

    def get_vulnerability_details(self, id):
        if id is None:
            return [False, "No vulnerability ID provided"]

        url = f"{self.url}/api/scanning/v1/anchore/query/vulnerabilities"

        params = {
                "id": id,
        }

        res = self.http.get(url, params=params, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        json_res = res.json()
        if "vulnerabilities" not in json_res or not json_res["vulnerabilities"]:
            return [False, f"Vulnerability {id} was not found"]

        return [True, json_res["vulnerabilities"][0]]

    def add_vulnerability_exception_bundle(self, name, comment=""):
        if not name:
            return [False, "A name is required for the exception bundle"]

        url = f"{self.url}/api/scanning/v1/vulnexceptions"
        params = {
                "version": "1_0",
                "name":    name,
                "comment": comment,
        }

        data = json.dumps(params)
        res = self.http.post(url, headers=self.hdrs, data=data, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def delete_vulnerability_exception_bundle(self, id):

        url = self.url + f"/api/scanning/v1/vulnexceptions/{id}"

        res = self.http.delete(url, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, None]

    def list_vulnerability_exception_bundles(self):
        url = f"{self.url}/api/scanning/v1/vulnexceptions"

        params = {
                "bundleId": "default",
        }

        res = self.http.get(url, params=params, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.json()]

    def get_vulnerability_exception_bundle(self, bundle):
        url = f"{self.url}/api/scanning/v1/vulnexceptions/{bundle}"

        params = {
                "bundleId": "default",
        }

        res = self.http.get(url, params=params, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        res_json = res.json()
        for item in res_json["items"]:
            item["trigger_id"] = str(item["trigger_id"]).rstrip("+*")
        return [True, res_json]

    def add_vulnerability_exception(self, bundle, cve, note=None, expiration_date=None):
        url = f"{self.url}/api/scanning/v1/vulnexceptions/{bundle}/vulnerabilities"

        params = {
                "gate":            "vulnerabilities",
                "is_busy":         False,
                "trigger_id":      f"{cve}+*",
                "expiration_date": int(expiration_date) if expiration_date else None,
                "notes":           note,
        }

        data = json.dumps(params)
        res = self.http.post(url, headers=self.hdrs, data=data, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        res_json = res.json()
        res_json["trigger_id"] = str(res_json["trigger_id"]).rstrip("+*")
        return [True, res_json]

    def delete_vulnerability_exception(self, bundle, id):
        url = f"{self.url}/api/scanning/v1/vulnexceptions/{bundle}/vulnerabilities/{id}"

        params = {
                "bundleId": "default",
        }

        res = self.http.delete(url, params=params, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, None]

    def update_vulnerability_exception(self, bundle, id, cve, enabled, note, expiration_date):
        url = f"{self.url}/api/scanning/v1/vulnexceptions/{bundle}/vulnerabilities/{id}"

        data = {
                "id":              id,
                "gate":            "vulnerabilities",
                "trigger_id":      f"{cve}+*",
                "enabled":         enabled,
                "notes":           note,
                "expiration_date": int(expiration_date) if expiration_date else None,
        }
        params = {
                "bundleId": "default",
        }

        res = self.http.put(url, data=json.dumps(data), params=params, headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        res_json = res.json()
        res_json["trigger_id"] = str(res_json["trigger_id"]).rstrip("+*")
        return [True, res_json]

    def download_cve_report_csv(self, vuln_type="os", scope_type="static"):
        """
        Downloads a CVE report in CSV format

        Args:
            vuln_type (str): Vulnerability type, can be either "os" or "non-os".
            scope_type (str): Scope type. Can be either "static" or "runtime".

        Returns:
            A tuple of (bool, str).
            The first parameter, if true, means that the result is correct, while
            if false, means that there's been an error. The second parameter
            will hold the response of the API call.
        """
        url = f"{self.url}/api/scanning/v1/reports/csv"

        params = {
                "queryType":        "vuln",
                "scopeType":        scope_type,
                "staticScope":      {
                        "registry":   "",
                        "repository": "",
                        "tag":        ""
                },
                "runtimeScope":     {},
                "imageQueryFilter": {"vType": vuln_type},
                "offset":           0,
                "limit":            100000
        }

        res = self.http.post(url, data=json.dumps(params), headers=self.hdrs, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        return [True, res.content.decode("utf-8")]

    def get_image_scanning_results(self, image_name, policy_id=None):
        '''
        Args:
            image_name (str): Image name to retrieve the scanning results from
            policy_id (str): Policy ID to check against. If not specified, will check against all policies.

        Returns:
            A tuple of (bool, str).
            The first parameter, if true, means that the result is correct, while
            if false, means that there's been an error. The second parameter
            will hold the response of the API call.
        '''
        try:
            ok, res = self.get_image(image_name)
            if not ok:
                return ok, res

            image_digest = res[0]["imageDigest"]
            image_tag = res[0]["image_detail"][0]["fulltag"]
        except RetryError:
            return [False, "could not retrieve image digest for the given image name, "
                           "ensure that the image has been scanned"]

        url = f"{self.url}/api/scanning/v1/images/{image_digest}/policyEvaluation"
        params = {
                "tag": image_tag,
        }

        res = self.http.get(url, headers=self.hdrs, params=params, verify=self.ssl_verify)
        if not self._checkResponse(res):
            return [False, self.lasterr]

        json_res = res.json()

        result = {
                "image_digest":    json_res["imageDigest"],
                "image_id":        json_res["imageId"],
                "status":          json_res["status"],
                "image_tag":       image_tag,
                "total_stop":      json_res["nStop"],
                "total_warn":      json_res["nWarn"],
                "last_evaluation": datetime.utcfromtimestamp(json_res["at"]),
                "policy_id":       "*",
                "policy_name":     "All policies",
                "warn_results":    [],
                "stop_results":    []
        }

        if policy_id:
            policy_results = [result for result in json_res["results"] if result["policyId"] == policy_id]
            if policy_results:
                filtered_result_by_policy_id = policy_results[0]
                result["policy_id"] = filtered_result_by_policy_id["policyId"]
                result["policy_name"] = filtered_result_by_policy_id["policyName"]
                result["total_stop"] = filtered_result_by_policy_id["nStop"]
                result["total_warn"] = filtered_result_by_policy_id["nWarn"]
                result["warn_results"] = [rule_result["checkOutput"]
                                          for gate_result in filtered_result_by_policy_id["gateResults"]
                                          for rule_result in gate_result["ruleResults"]
                                          if rule_result["gateAction"] == "warn"]
                result["stop_results"] = [rule_result["checkOutput"]
                                          for gate_result in filtered_result_by_policy_id["gateResults"]
                                          for rule_result in gate_result["ruleResults"]
                                          if rule_result["gateAction"] == "stop"]
            else:
                return [False, "the specified policy ID doesn't exist"]
        else:
            result["warn_results"] = [rule_result["checkOutput"]
                                      for result in json_res["results"]
                                      for gate_result in result["gateResults"]
                                      for rule_result in gate_result["ruleResults"]
                                      if rule_result["gateAction"] == "warn"]
            result["stop_results"] = [rule_result["checkOutput"]
                                      for result in json_res["results"]
                                      for gate_result in result["gateResults"]
                                      for rule_result in gate_result["ruleResults"]
                                      if rule_result["gateAction"] == "stop"]

        return [True, result]
